# SPDX-License-Identifier: MIT
import inspect
import os
import unittest
from xml.etree import ElementTree

import jinja2
from packaging.version import Version

import odxtools_porsche
from odxtools.compumethods.compucategory import CompuCategory
from odxtools.compumethods.compucodecompumethod import CompuCodeCompuMethod
from odxtools.compumethods.compuconst import CompuConst
from odxtools.compumethods.compuinternaltophys import CompuInternalToPhys
from odxtools.compumethods.compuphystointernal import CompuPhysToInternal
from odxtools.compumethods.compurationalcoeffs import CompuRationalCoeffs
from odxtools.compumethods.compuscale import CompuScale
from odxtools.compumethods.createanycompumethod import create_any_compu_method_from_et
from odxtools.compumethods.intervaltype import IntervalType
from odxtools.compumethods.limit import Limit
from odxtools.compumethods.linearcompumethod import LinearCompuMethod
from odxtools.compumethods.ratfunccompumethod import RatFuncCompuMethod
from odxtools.compumethods.scaleratfunccompumethod import ScaleRatFuncCompuMethod
from odxtools.compumethods.tabintpcompumethod import TabIntpCompuMethod
from odxtools.exceptions import DecodeError, EncodeError, OdxError
from odxtools.odxdoccontext import OdxDocContext
from odxtools.odxlink import DocType, OdxDocFragment
from odxtools.odxtypes import DataType
from odxtools.progcode import ProgCode
from odxtools.writepdxfile import (get_parent_container_name, jinja2_odxraise_helper,
                                   make_bool_xml_attrib, make_xml_attrib)

doc_frags = [OdxDocFragment("UnitTest", DocType.CONTAINER)]


class TestLinearCompuMethod(unittest.TestCase):

    def setUp(self) -> None:
        """Prepares the jinja environment and the sample linear compumethod"""

        def _get_jinja_environment() -> jinja2.environment.Environment:
            __module_filename = inspect.getsourcefile(odxtools_porsche)
            assert isinstance(__module_filename, str)
            templates_dir = os.path.sep.join([os.path.dirname(__module_filename), "templates"])

            jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(templates_dir))

            jinja_env.globals["getattr"] = getattr
            jinja_env.globals["hasattr"] = hasattr
            jinja_env.globals["odxraise"] = jinja2_odxraise_helper
            jinja_env.globals["make_xml_attrib"] = make_xml_attrib
            jinja_env.globals["make_bool_xml_attrib"] = make_bool_xml_attrib
            jinja_env.globals["get_parent_container_name"] = get_parent_container_name

            return jinja_env

        self.jinja_env = _get_jinja_environment()

        self.linear_compumethod = LinearCompuMethod(
            category=CompuCategory.LINEAR,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_INT32,
                        numerators=[0, 1],
                        denominators=[3600],
                    ),
                    domain_type=DataType.A_INT32,
                    range_type=DataType.A_INT32),
            ]),
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_INT32,
        )

        self.linear_compumethod_xml = f"""
        <COMPU-METHOD>
            <CATEGORY>LINEAR</CATEGORY>
            <COMPU-INTERNAL-TO-PHYS>
                <COMPU-SCALES>
                    <COMPU-SCALE>
                        <COMPU-RATIONAL-COEFFS>
                            <COMPU-NUMERATOR>
                                <V>{self.linear_compumethod.segment.offset}</V>
                                <V>{self.linear_compumethod.segment.factor}</V>
                            </COMPU-NUMERATOR>
                            <COMPU-DENOMINATOR>
                                <V>{self.linear_compumethod.segment.denominator}</V>
                            </COMPU-DENOMINATOR>
                        </COMPU-RATIONAL-COEFFS>
                    </COMPU-SCALE>
                </COMPU-SCALES>
            </COMPU-INTERNAL-TO-PHYS>
        </COMPU-METHOD>
        """

    def test_read_odx(self) -> None:
        """Test parsing of linear compumethod"""
        expected = self.linear_compumethod

        et_element = ElementTree.fromstring(self.linear_compumethod_xml)
        actual = create_any_compu_method_from_et(
            et_element,
            OdxDocContext(Version("2.2.0"), doc_frags),
            internal_type=expected.internal_type,
            physical_type=expected.physical_type)
        self.assertIsInstance(actual, LinearCompuMethod)
        assert isinstance(actual, LinearCompuMethod)
        self.assertEqual(expected.physical_type, actual.physical_type)
        self.assertEqual(expected.internal_type, actual.internal_type)
        self.assertEqual(expected.segment.offset, actual.segment.offset)
        self.assertEqual(expected.segment.factor, actual.segment.factor)
        self.assertEqual(expected.segment.denominator, actual.segment.denominator)

    def test_write_odx(self) -> None:
        self.maxDiff = None
        dlc_tpl = self.jinja_env.get_template("macros/printCompuMethod.xml.jinja2")
        module = dlc_tpl.make_module()

        actual_xml = module.printCompuMethod(self.linear_compumethod)  # type: ignore[attr-defined]

        expected_xml = self.linear_compumethod_xml

        # We ignore spaces
        def remove_spaces(string: str) -> str:
            return "".join(string.split())

        self.assertEqual(remove_spaces(actual_xml), remove_spaces(expected_xml))

    def test_linear_compu_method_type_denom_not_one(self) -> None:
        compu_method = LinearCompuMethod(
            category=CompuCategory.LINEAR,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    lower_limit=Limit(
                        value_raw="0",
                        value_type=DataType.A_INT32,
                        interval_type=IntervalType.INFINITE),
                    upper_limit=Limit(
                        value_raw="0",
                        value_type=DataType.A_INT32,
                        interval_type=IntervalType.INFINITE),
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_INT32,
                        numerators=[0, 1],
                        denominators=[3600],
                    ),
                    domain_type=DataType.A_INT32,
                    range_type=DataType.A_INT32),
            ]),
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_INT32)

        self.assertEqual(compu_method.convert_physical_to_internal(2), 7200)
        self.assertEqual(compu_method.convert_internal_to_physical(7200), 2)

    def test_linear_compu_method_type_int_int(self) -> None:
        compu_method = LinearCompuMethod(
            category=CompuCategory.LINEAR,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_INT32,
                        numerators=[1, 3],
                        denominators=[1],
                    ),
                    domain_type=DataType.A_INT32,
                    range_type=DataType.A_INT32),
            ]),
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_INT32)

        self.assertEqual(compu_method.convert_internal_to_physical(4), 13)
        self.assertEqual(compu_method.convert_internal_to_physical(0), 1)
        self.assertEqual(compu_method.convert_internal_to_physical(-2), -5)

        self.assertEqual(compu_method.convert_physical_to_internal(13), 4)
        self.assertEqual(compu_method.convert_physical_to_internal(1), 0)
        self.assertEqual(compu_method.convert_physical_to_internal(-5), -2)

    def test_linear_compu_method_type_int_float(self) -> None:
        compu_method = LinearCompuMethod(
            category=CompuCategory.LINEAR,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_INT32,
                        numerators=[1, 3],
                        denominators=[1],
                    ),
                    domain_type=DataType.A_INT32,
                    range_type=DataType.A_FLOAT32),
            ]),
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_FLOAT32)

        self.assertTrue(compu_method.is_valid_internal_value(123))
        self.assertFalse(compu_method.is_valid_internal_value("123"))
        self.assertFalse(compu_method.is_valid_internal_value(1.2345))

        self.assertTrue(compu_method.is_valid_physical_value(1.2345))
        self.assertTrue(compu_method.is_valid_physical_value(123))
        self.assertFalse(compu_method.is_valid_physical_value("123"))

    def test_linear_compu_method_type_float_int(self) -> None:
        compu_method = LinearCompuMethod(
            category=CompuCategory.LINEAR,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_INT32,
                        numerators=[1, 3],
                        denominators=[1],
                    ),
                    domain_type=DataType.A_INT32,
                    range_type=DataType.A_INT32),
            ]),
            internal_type=DataType.A_FLOAT32,
            physical_type=DataType.A_INT32)
        self.assertTrue(compu_method.is_valid_internal_value(1.2345))
        self.assertTrue(compu_method.is_valid_internal_value(123))
        self.assertFalse(compu_method.is_valid_internal_value("123"))

        self.assertTrue(compu_method.is_valid_physical_value(123))
        self.assertFalse(compu_method.is_valid_physical_value("123"))
        self.assertFalse(compu_method.is_valid_physical_value(1.2345))

    def test_linear_compu_method_type_string(self) -> None:
        with self.assertRaises(OdxError):
            LinearCompuMethod(
                category=CompuCategory.LINEAR,
                compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                    CompuScale(
                        compu_rational_coeffs=CompuRationalCoeffs(
                            value_type=DataType.A_INT32,
                            numerators=[1, 3],
                            denominators=[1],
                        ),
                        domain_type=DataType.A_INT32,
                        range_type=DataType.A_INT32),
                ]),
                internal_type=DataType.A_ASCIISTRING,
                physical_type=DataType.A_UNICODE2STRING)

    def test_linear_compu_method_limits(self) -> None:
        compu_method = LinearCompuMethod(
            category=CompuCategory.LINEAR,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    lower_limit=Limit(value_raw="2", value_type=DataType.A_INT32),
                    upper_limit=Limit(value_raw="15", value_type=DataType.A_INT32),
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_INT32,
                        numerators=[1, 5],
                        denominators=[1],
                    ),
                    domain_type=DataType.A_INT32,
                    range_type=DataType.A_INT32),
            ]),
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_INT32)

        self.assertFalse(compu_method.is_valid_internal_value(-3))
        self.assertFalse(compu_method.is_valid_internal_value(1))
        self.assertFalse(compu_method.is_valid_internal_value(16))

        self.assertTrue(compu_method.is_valid_internal_value(2))
        self.assertTrue(compu_method.is_valid_internal_value(15))
        self.assertTrue(compu_method.is_valid_internal_value(7))

        self.assertFalse(compu_method.is_valid_physical_value(10))
        self.assertFalse(compu_method.is_valid_physical_value(77))

        self.assertTrue(compu_method.is_valid_physical_value(11))
        self.assertTrue(compu_method.is_valid_physical_value(21))
        self.assertTrue(compu_method.is_valid_physical_value(76))

        self.assertEqual(compu_method.convert_internal_to_physical(4), 21)
        self.assertEqual(compu_method.convert_physical_to_internal(21), 4)

    def test_linear_compu_method_physical_limits(self) -> None:
        # Define decoding function: f: (2, 15] -> [-74, -14], f(x) = -5*x + 1
        compu_method = LinearCompuMethod(
            category=CompuCategory.LINEAR,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    lower_limit=Limit(
                        value_raw="2", value_type=DataType.A_INT32,
                        interval_type=IntervalType.OPEN),
                    upper_limit=Limit(value_raw="15", value_type=DataType.A_INT32),
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_INT32,
                        numerators=[1, -5],
                        denominators=[1],
                    ),
                    domain_type=DataType.A_INT32,
                    range_type=DataType.A_INT32),
            ]),
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_INT32)

        assert compu_method.segment.internal_lower_limit is not None
        assert compu_method.segment.internal_upper_limit is not None
        assert compu_method.segment.physical_lower_limit is not None
        assert compu_method.segment.physical_upper_limit is not None

        self.assertEqual(
            compu_method.segment.internal_lower_limit,
            Limit(value_raw="2", value_type=DataType.A_INT32, interval_type=IntervalType.OPEN))
        self.assertEqual(compu_method.segment.internal_upper_limit,
                         Limit(value_raw="15", value_type=DataType.A_INT32))
        self.assertEqual(compu_method.segment.internal_upper_limit.interval_type, None)

        self.assertEqual(compu_method.segment.physical_lower_limit,
                         Limit(value_raw="-74", value_type=DataType.A_INT32))
        self.assertEqual(
            compu_method.segment.physical_upper_limit,
            Limit(value_raw="-9", value_type=DataType.A_INT32, interval_type=IntervalType.OPEN))

        self.assertFalse(compu_method.segment.internal_lower_limit.complies_to_lower(2))
        self.assertTrue(compu_method.segment.internal_lower_limit.complies_to_lower(3))
        self.assertTrue(compu_method.segment.internal_upper_limit.complies_to_upper(15))
        self.assertFalse(compu_method.segment.internal_upper_limit.complies_to_upper(16))

        self.assertFalse(compu_method.segment.physical_lower_limit.complies_to_lower(-75))
        self.assertTrue(compu_method.segment.physical_lower_limit.complies_to_lower(-74))
        self.assertTrue(compu_method.segment.physical_upper_limit.complies_to_upper(-10))
        self.assertFalse(compu_method.segment.physical_upper_limit.complies_to_upper(-9))

        self.assertTrue(compu_method.is_valid_internal_value(3))
        self.assertTrue(compu_method.is_valid_internal_value(15))
        self.assertFalse(compu_method.is_valid_internal_value(2))
        self.assertFalse(compu_method.is_valid_internal_value(16))

        self.assertFalse(compu_method.is_valid_physical_value(-75))
        self.assertTrue(compu_method.is_valid_physical_value(-74))
        self.assertTrue(compu_method.is_valid_physical_value(-14))
        self.assertTrue(compu_method.is_valid_physical_value(-10))
        self.assertFalse(compu_method.is_valid_physical_value(-9))


class TestCompuCodeCompuMethod(unittest.TestCase):

    def test_compu_code_compu_method(self) -> None:
        compu_method = CompuCodeCompuMethod(
            category=CompuCategory.COMPUCODE,
            compu_internal_to_phys=CompuInternalToPhys(
                prog_code=ProgCode(
                    code_file="nice_computation.java", syntax="JAVA", revision="1.0")),
            compu_phys_to_internal=CompuPhysToInternal(
                prog_code=ProgCode(
                    code_file="nice_inverse_computation.java", syntax="JAVA", revision="1.0")),
            internal_type=DataType.A_FLOAT32,
            physical_type=DataType.A_FLOAT32)

        # COMPUCODE compu methods can only be used from Java projects
        # (i.e., not odxtools)
        with self.assertRaises(DecodeError):
            compu_method.convert_internal_to_physical(1)
        with self.assertRaises(EncodeError):
            compu_method.convert_physical_to_internal(2)


class TestRatFuncCompuMethod(unittest.TestCase):

    def test_rat_func_compu_method(self) -> None:
        compu_method = RatFuncCompuMethod(
            category=CompuCategory.RAT_FUNC,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    lower_limit=Limit(
                        value_raw="2",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    upper_limit=Limit(
                        value_raw="3",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_FLOAT32,
                        # f(x) = 3.14*(2*x^2 + 4x + 6) / 3.14
                        numerators=[3.14 * 6, 3.14 * 4, 3.14 * 2],
                        denominators=[3.14],
                    ),
                    domain_type=DataType.A_FLOAT32,
                    range_type=DataType.A_FLOAT32),
            ]),
            internal_type=DataType.A_FLOAT32,
            physical_type=DataType.A_FLOAT32)

        phys_val = compu_method.convert_internal_to_physical(2.5)
        assert isinstance(phys_val, float)
        self.assertTrue(abs(phys_val - 28.5) < 1e-5)

        # out of range
        with self.assertRaises(DecodeError):
            compu_method.convert_internal_to_physical(3.01)
        with self.assertRaises(DecodeError):
            compu_method.convert_internal_to_physical(1.99)

        # if the inverse function is not explicitly specified,
        # inversion is not allowed!
        with self.assertRaises(EncodeError):
            compu_method.convert_physical_to_internal(2.5)

    def test_rat_func_compu_method_with_inverse(self) -> None:
        compu_method = RatFuncCompuMethod(
            category=CompuCategory.RAT_FUNC,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    lower_limit=Limit(
                        value_raw="2",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    upper_limit=Limit(
                        value_raw="3",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_FLOAT32,
                        # f(x) = x^2
                        numerators=[0, 0, 1],
                        denominators=[1],
                    ),
                    domain_type=DataType.A_FLOAT32,
                    range_type=DataType.A_FLOAT32),
            ]),
            compu_phys_to_internal=CompuPhysToInternal(compu_scales=[
                CompuScale(
                    lower_limit=Limit(
                        value_raw="4",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    upper_limit=Limit(
                        value_raw="9",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_FLOAT32,
                        # first three terms of Taylor expansion of
                        # f(x) = sqrt(x) at evaluation point 1
                        numerators=[2.72 * 3 / 8, 2.72 * 3 / 4, 2.72 * -1 / 8],
                        denominators=[2.72],
                    ),
                    domain_type=DataType.A_FLOAT32,
                    range_type=DataType.A_FLOAT32),
            ]),
            internal_type=DataType.A_FLOAT32,
            physical_type=DataType.A_FLOAT32)

        self.assertEqual(compu_method.convert_internal_to_physical(2), 4)
        self.assertEqual(compu_method.convert_internal_to_physical(2.5), 6.25)
        phys_val = compu_method.convert_internal_to_physical(3)
        assert isinstance(phys_val, float)
        self.assertTrue(abs(phys_val - 9) < 1e-8)

        # note that the inverse values are pretty inaccurate because
        # the Taylor series was cut off quite early.
        int_val = compu_method.convert_physical_to_internal(4)
        assert isinstance(int_val, float)
        self.assertTrue(abs(int_val - 1.375) < 1e-4)
        int_val = compu_method.convert_physical_to_internal(6.25)
        assert isinstance(int_val, float)
        self.assertTrue(abs(int_val - 0.17969) < 1e-4)
        self.assertEqual(compu_method.convert_physical_to_internal(9), -3)

        # ensure that we stay in bounds
        with self.assertRaises(DecodeError):
            compu_method.convert_internal_to_physical(3.99)
        with self.assertRaises(DecodeError):
            compu_method.convert_internal_to_physical(9.01)


class TestScaleRatFuncCompuMethod(unittest.TestCase):

    def test_scale_rat_func_compu_method(self) -> None:
        compu_method = ScaleRatFuncCompuMethod(
            category=CompuCategory.SCALE_RAT_FUNC,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    lower_limit=Limit(
                        value_raw="2",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    upper_limit=Limit(
                        value_raw="3",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_FLOAT32,
                        # f(x) = 3.14*(2*x^2 + 4x + 6) / 3.14
                        numerators=[3.14 * 6, 3.14 * 4, 3.14 * 2],
                        denominators=[3.14],
                    ),
                    domain_type=DataType.A_FLOAT32,
                    range_type=DataType.A_FLOAT32),
            ]),
            internal_type=DataType.A_FLOAT32,
            physical_type=DataType.A_FLOAT32)

        phys_val = compu_method.convert_internal_to_physical(2.5)
        assert isinstance(phys_val, float)
        self.assertTrue(abs(phys_val - 28.5) < 1e-5)

        # out of range
        with self.assertRaises(DecodeError):
            compu_method.convert_internal_to_physical(3.01)
        with self.assertRaises(DecodeError):
            compu_method.convert_internal_to_physical(1.99)

        # if the inverse function is not explicitly specified,
        # inversion is not allowed!
        with self.assertRaises(EncodeError):
            compu_method.convert_physical_to_internal(2.5)

    def test_scale_rat_func_compu_method_with_inverse(self) -> None:
        compu_method = ScaleRatFuncCompuMethod(
            category=CompuCategory.SCALE_RAT_FUNC,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    lower_limit=Limit(
                        value_raw="2",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    upper_limit=Limit(
                        value_raw="3",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_FLOAT32,
                        # f(x) = x^2
                        numerators=[0, 0, 1],
                        denominators=[1],
                    ),
                    domain_type=DataType.A_FLOAT32,
                    range_type=DataType.A_FLOAT32),
                CompuScale(
                    lower_limit=Limit(
                        value_raw="4",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    upper_limit=Limit(
                        value_raw="5",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_FLOAT32,
                        # f(x) = 1.38e-23*(3*x + 10)/1.38e-23
                        numerators=[1.38e-23 * 10, 1.38e-23 * 3],
                        denominators=[1.38e-23],
                    ),
                    domain_type=DataType.A_FLOAT32,
                    range_type=DataType.A_FLOAT32),
            ]),
            compu_phys_to_internal=CompuPhysToInternal(compu_scales=[
                CompuScale(
                    lower_limit=Limit(
                        value_raw="4",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    upper_limit=Limit(
                        value_raw="9",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_FLOAT32,
                        # first three terms of Taylor expansion of
                        # f(x) = sqrt(x) at evaluation point 1
                        numerators=[2.72 * 3 / 8, 2.72 * 3 / 4, 2.72 * -1 / 8],
                        denominators=[2.72],
                    ),
                    domain_type=DataType.A_FLOAT32,
                    range_type=DataType.A_FLOAT32),
                CompuScale(
                    lower_limit=Limit(
                        value_raw="22",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    upper_limit=Limit(
                        value_raw="25",
                        value_type=DataType.A_FLOAT32,
                        interval_type=IntervalType.CLOSED),
                    compu_rational_coeffs=CompuRationalCoeffs(
                        value_type=DataType.A_FLOAT32,
                        # f(x) = (x - 10)/3
                        numerators=[-10, 1],
                        denominators=[3],
                    ),
                    domain_type=DataType.A_FLOAT32,
                    range_type=DataType.A_FLOAT32),
            ]),
            internal_type=DataType.A_FLOAT32,
            physical_type=DataType.A_FLOAT32)

        self.assertEqual(compu_method.convert_internal_to_physical(2), 4)
        self.assertEqual(compu_method.convert_internal_to_physical(2.5), 6.25)
        phys_val = compu_method.convert_internal_to_physical(3)
        assert isinstance(phys_val, float)
        self.assertTrue(abs(phys_val - 9) < 1e-8)

        phys_val = compu_method.convert_internal_to_physical(4)
        assert isinstance(phys_val, float)
        self.assertTrue(abs(phys_val - 22) < 1e-8)
        int_val = compu_method.convert_physical_to_internal(22)
        assert isinstance(int_val, float)
        self.assertTrue(abs(int_val - 4) < 1e-8)

        # note that the inverse values are pretty inaccurate because
        # the Taylor series was cut off way quite early.
        int_val = compu_method.convert_physical_to_internal(4)
        assert isinstance(int_val, float)
        self.assertTrue(abs(int_val - 1.375) < 1e-4)
        int_val = compu_method.convert_physical_to_internal(6.25)
        assert isinstance(int_val, float)
        self.assertTrue(abs(int_val - 0.17969) < 1e-4)
        self.assertEqual(compu_method.convert_physical_to_internal(9), -3)

        # make sure that we stay in bounds
        with self.assertRaises(DecodeError):
            compu_method.convert_internal_to_physical(3.99)
        with self.assertRaises(DecodeError):
            compu_method.convert_internal_to_physical(9.01)


class TestTabIntpCompuMethod(unittest.TestCase):

    def setUp(self) -> None:
        """Prepares the jinja environment and the sample tab-intp compumethod"""

        def _get_jinja_environment() -> jinja2.environment.Environment:
            __module_filename = inspect.getsourcefile(odxtools_porsche)
            assert isinstance(__module_filename, str)
            templates_dir = os.path.sep.join([os.path.dirname(__module_filename), "templates"])

            jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(templates_dir))

            # allows to put XML attributes on a separate line while it is
            # collapsed with the previous line in the rendering
            jinja_env.filters["odxtools_collapse_xml_attribute"] = (lambda x: " " + x.strip()
                                                                    if x.strip() else "")

            jinja_env.globals["getattr"] = getattr
            jinja_env.globals["hasattr"] = hasattr
            return jinja_env

        self.jinja_env = _get_jinja_environment()

        self.tab_intp_compumethod = TabIntpCompuMethod(
            category=CompuCategory.TAB_INTP,
            compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                CompuScale(
                    lower_limit=Limit(value_raw="0", value_type=DataType.A_INT32),
                    compu_const=CompuConst(v="-1", data_type=DataType.A_INT32),
                    domain_type=DataType.A_INT32,
                    range_type=DataType.A_INT32),
                CompuScale(
                    lower_limit=Limit(value_raw="10", value_type=DataType.A_INT32),
                    compu_const=CompuConst(v="1", data_type=DataType.A_INT32),
                    domain_type=DataType.A_INT32,
                    range_type=DataType.A_INT32),
                CompuScale(
                    lower_limit=Limit(value_raw="30", value_type=DataType.A_INT32),
                    compu_const=CompuConst(v="2", data_type=DataType.A_INT32),
                    domain_type=DataType.A_INT32,
                    range_type=DataType.A_INT32),
            ]),
            internal_type=DataType.A_INT32,
            physical_type=DataType.A_FLOAT32,
        )

        self.tab_intp_compumethod_xml = f"""
        <COMPU-METHOD>
            <CATEGORY>TAB-INTP</CATEGORY>
            <COMPU-INTERNAL-TO-PHYS>
                <COMPU-SCALES>
                    <COMPU-SCALE>
                        <LOWER-LIMIT>{self.tab_intp_compumethod.internal_points[0]}</LOWER-LIMIT>
                        <COMPU-CONST>
                            <V>{self.tab_intp_compumethod.physical_points[0]}</V>
                        </COMPU-CONST>
                    </COMPU-SCALE>
                    <COMPU-SCALE>
                        <LOWER-LIMIT>{self.tab_intp_compumethod.internal_points[1]}</LOWER-LIMIT>
                        <COMPU-CONST>
                            <V>{self.tab_intp_compumethod.physical_points[1]}</V>
                        </COMPU-CONST>
                    </COMPU-SCALE>
                    <COMPU-SCALE>
                        <LOWER-LIMIT>{self.tab_intp_compumethod.internal_points[2]}</LOWER-LIMIT>
                        <COMPU-CONST>
                            <V>{self.tab_intp_compumethod.physical_points[2]}</V>
                        </COMPU-CONST>
                    </COMPU-SCALE>
                </COMPU-SCALES>
            </COMPU-INTERNAL-TO-PHYS>
        </COMPU-METHOD>
        """

    def test_tabintp_convert_type_int_float(self) -> None:
        method = self.tab_intp_compumethod

        for internal, physical in [
            (0, -1),
            (2, -0.6),
            (3, -0.4),
            (5, 0),
            (10, 1),
            (20, 1.5),
            (25, 1.75),
            (30, 2),
        ]:
            self.assertTrue(method.is_valid_internal_value(internal))
            self.assertTrue(method.is_valid_physical_value(physical))
            self.assertEqual(method.convert_internal_to_physical(internal), physical)
            self.assertEqual(method.convert_physical_to_internal(physical), internal)

        self.assertRaises(DecodeError, method.convert_internal_to_physical, -2)
        self.assertRaises(DecodeError, method.convert_internal_to_physical, 31)
        self.assertRaises(EncodeError, method.convert_physical_to_internal, -2)
        self.assertRaises(EncodeError, method.convert_physical_to_internal, 2.1)

    def test_read_odx(self) -> None:
        expected = self.tab_intp_compumethod

        et_element = ElementTree.fromstring(self.tab_intp_compumethod_xml)
        actual = create_any_compu_method_from_et(
            et_element,
            OdxDocContext(Version("2.2.0"), doc_frags),
            internal_type=expected.internal_type,
            physical_type=expected.physical_type)
        self.assertIsInstance(actual, TabIntpCompuMethod)
        assert isinstance(expected, TabIntpCompuMethod)
        assert isinstance(actual, TabIntpCompuMethod)
        self.assertEqual(expected.physical_type, actual.physical_type)
        self.assertEqual(expected.internal_type, actual.internal_type)
        self.assertEqual(expected.internal_points, actual.internal_points)
        self.assertEqual(expected.physical_points, actual.physical_points)

    def test_write_odx(self) -> None:
        dlc_tpl = self.jinja_env.get_template("macros/printCompuMethod.xml.jinja2")
        module = dlc_tpl.make_module()

        actual_xml = module.printCompuMethod(  # type: ignore[attr-defined]
            self.tab_intp_compumethod)

        expected_xml = self.tab_intp_compumethod_xml

        # We ignore spaces
        def remove_spaces(string: str) -> str:
            return "".join(string.split())

        self.assertEqual(remove_spaces(actual_xml), remove_spaces(expected_xml))


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: MIT
import argparse
from collections.abc import Callable

import rich

from ..comparaminstance import ComparamInstance
from ..database import Database
from ..dataobjectproperty import DataObjectProperty
from ..diagcomm import DiagComm
from ..diaglayers.basevariant import BaseVariant
from ..diaglayers.diaglayer import DiagLayer
from ..diaglayers.ecuvariant import EcuVariant
from ..diaglayers.hierarchyelement import HierarchyElement
from ..diagservice import DiagService
from ..singleecujob import SingleEcuJob
from . import _parser_utils
from ._parser_utils import SubparsersList
from ._print_utils import format_desc, print_diagnostic_service, print_dl_metrics

# name of the tool
_odxtools_tool_name_ = "list"


def print_summary(odxdb: Database,
                  print_global_negative_responses: bool = False,
                  print_services: bool = False,
                  print_dops: bool = False,
                  print_params: bool = False,
                  print_comparams: bool = False,
                  print_pre_condition_states: bool = False,
                  print_state_transitions: bool = False,
                  print_audiences: bool = False,
                  allow_unknown_bit_lengths: bool = False,
                  variants: list[str] | None = None,
                  service_filter: Callable[[DiagComm], bool] = lambda x: True) -> None:

    diag_layer_names = [dl.short_name for dl in odxdb.diag_layers]
    diag_layers: list[DiagLayer] = []

    if variants is None:
        variants = diag_layer_names

    for name in variants:
        if name in diag_layer_names:
            diag_layers.append([x for x in odxdb.diag_layers if x.short_name == name][0])

        else:
            rich.print(f"The variant '{name}' could not be found!")
            return

    if diag_layers:
        rich.print("\n")
        rich.print(f"Overview of diagnostic layers: ")
        print_dl_metrics(diag_layers)

    for dl in diag_layers:
        rich.print("\n")
        rich.print(f"Diagnostic layer: '{dl.short_name}'")
        rich.print(f" Variant Type: {dl.variant_type.value}")

        all_services: list[DiagComm] = sorted(dl.services, key=lambda x: x.short_name)

        if isinstance(dl, (BaseVariant, EcuVariant)):
            for proto in dl.protocols:
                if (can_rx_id := dl.get_can_receive_id(proto.short_name)) is not None:
                    rich.print(
                        f"  CAN receive ID for protocol '{proto.short_name}': 0x{can_rx_id:x}")

                if (can_tx_id := dl.get_can_send_id(proto.short_name)) is not None:
                    rich.print(f"  CAN send ID for protocol '{proto.short_name}': 0x{can_tx_id:x}")

        if dl.description:
            desc = format_desc(dl.description, indent=2)
            rich.print(f" Description: " + desc)

        if print_global_negative_responses and dl.global_negative_responses:
            rich.print("\n")
            rich.print(f"The global negative responses of '{dl.short_name}' are: ")
            for gnr in dl.global_negative_responses:
                rich.print(f" {gnr.short_name}")

        if print_services and len(all_services) > 0:
            services = [s for s in all_services if service_filter(s)]
            if len(services) > 0:
                rich.print("\n")
                rich.print(f"The services of '{dl.short_name}' are: ")
                for service in services:
                    if isinstance(service, DiagService):
                        print_diagnostic_service(
                            service,
                            print_params=print_params,
                            print_pre_condition_states=print_pre_condition_states,
                            print_state_transitions=print_state_transitions,
                            print_audiences=print_audiences,
                            allow_unknown_bit_lengths=allow_unknown_bit_lengths)
                    elif isinstance(service, SingleEcuJob):
                        rich.print(f" Single ECU job: {service.odx_id}")
                    else:
                        rich.print(f" Unidentifiable service: {service}")
        ddd_spec = dl.diag_data_dictionary_spec
        data_object_properties: list[
            DataObjectProperty] = [] if ddd_spec is None else ddd_spec.data_object_props
        if print_dops and len(data_object_properties) > 0:
            rich.print("\n")
            rich.print(f"The DOPs of the {dl.variant_type.value} '{dl.short_name}' are: ")
            for dop in sorted(
                    data_object_properties, key=lambda x: (type(x).__name__, x.short_name)):
                rich.print("  " + str(dop.short_name).replace("\n", "\n  "))

        comparam_refs: list[ComparamInstance] = []
        if isinstance(dl, HierarchyElement):
            comparam_refs = dl.comparam_refs

        if print_comparams and len(comparam_refs) > 0:
            rich.print("\n")
            rich.print(
                f"The communication parameters of the {dl.variant_type.value} '{dl.short_name}' are: "
            )
            for com_param in comparam_refs:
                rich.print(f"  {com_param.short_name}: {com_param.value}")


def add_subparser(subparsers: SubparsersList) -> None:
    parser = subparsers.add_parser(
        "list",
        description="\n".join([
            "List the content of automotive diagnostic files (*.pdx)",
            "",
            "Examples:",
            "  For displaying only the names of the diagnostic layers use:",
            "    odxtools list ./path/to/database.pdx",
            "  For displaying all content use:",
            "    odxtools list ./path/to/database.pdx --all",
            "  For more information use:",
            "    odxtools list -h",
        ]),
        help="Print a summary of automotive diagnostic files.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    _parser_utils.add_pdx_argument(parser)

    parser.add_argument(
        "-v",
        "--variants",
        nargs="+",
        metavar="VARIANT",
        required=False,
        help="Specifies which variants should be included.",
        default="all",
    )

    parser.add_argument(
        "-g",
        "--global-negative-responses",
        default=False,
        action="store_const",
        const=True,
        required=False,
        help="Print a list of the global negative responses for the selected ECUs.",
    )

    # The service option is None if option is not passed at all (-> do not print services). It is an empty list if --services is passed
    parser.add_argument(
        "-s",
        "--services",
        nargs="*",
        default=None,
        metavar="SERVICE",
        required=False,
        help="Print a list of diagnostic services specified in the pdx. \n" +
        "If no service names are specified, all services are printed.",
    )
    # Pretty print message format and list parameters
    parser.add_argument(
        "-p",
        "--params",
        default=False,
        action="store_const",
        const=True,
        required=False,
        help="Print a list of all parameters relevant for the selected items.\n",
    )
    parser.add_argument(
        "-d",
        "--dops",
        default=False,
        action="store_const",
        const=True,
        required=False,
        help="Print a list of all data object properties relevant for the selected items",
    )

    # Shortcut to just dump everything
    parser.add_argument(
        "-a",
        "--all",
        default=False,
        action="store_const",
        const=True,
        required=False,
        help="Print a list of all diagnostic services and DOPs specified in the pdx",
    )

    parser.add_argument(
        "--dump-database",
        action="store_true",
        required=False,
        help="Ignore all other parameters and print a comprehensive dump the full database "
        "instead of providing a pretty-printed summary",
    )


def run(args: argparse.Namespace) -> None:
    odxdb = _parser_utils.load_file(args)

    if args.dump_database:
        print(repr(odxdb))
        return

    def service_filter(s: DiagComm) -> bool:
        if args.services and len(args.services) > 0:
            return s.short_name in args.services
        return True

    variants = args.variants if args.variants else None
    print_summary(
        odxdb,
        print_global_negative_responses=args.all or args.global_negative_responses,
        print_services=args.all or args.params or args.services is not None,
        service_filter=service_filter,
        print_dops=args.all or args.dops,
        variants=None if variants == "all" else variants,
        print_params=args.all or args.params,
        print_comparams=args.all,
        print_pre_condition_states=args.all,
        print_state_transitions=args.all,
        print_audiences=args.all,
        allow_unknown_bit_lengths=args.all)



# SPDX-License-Identifier: MIT
from collections.abc import Iterable
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, cast
from xml.etree import ElementTree

from typing_extensions import override

from ..diagvariable import DiagVariable, HasDiagVariables
from ..dyndefinedspec import DynDefinedSpec
from ..ecuvariantpattern import EcuVariantPattern
from ..exceptions import odxassert
from ..nameditemlist import NamedItemList
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkRef
from ..parentref import ParentRef
from ..variablegroup import HasVariableGroups, VariableGroup
from .basevariant import BaseVariant
from .diaglayer import DiagLayer
from .ecuvariantraw import EcuVariantRaw
from .hierarchyelement import HierarchyElement


@dataclass(kw_only=True)
class EcuVariant(HierarchyElement):

    @property
    def ecu_variant_raw(self) -> EcuVariantRaw:
        return cast(EcuVariantRaw, self.diag_layer_raw)

    @property
    def diag_variables_raw(self) -> list[DiagVariable | OdxLinkRef]:
        return self.ecu_variant_raw.diag_variables_raw

    @property
    def dyn_defined_spec(self) -> DynDefinedSpec | None:
        return self.ecu_variant_raw.dyn_defined_spec

    @property
    def parent_refs(self) -> list[ParentRef]:
        return self.ecu_variant_raw.parent_refs

    @property
    def base_variant(self) -> BaseVariant | None:
        """Return the base variant for the ECU variant

        The ODX specification allows at a single base variant for each
        ECU variant, cf checker rule 50 of appendix B.2 of the
        specification document.

        """
        for pr in self.ecu_variant_raw.parent_refs:
            if isinstance(pr.layer, BaseVariant):
                return pr.layer

        return None

    @property
    def ecu_variant_patterns(self) -> list[EcuVariantPattern]:
        return self.ecu_variant_raw.ecu_variant_patterns

    #######
    # <properties subject to value inheritance>
    #######
    @property
    def diag_variables(self) -> NamedItemList[DiagVariable]:
        return self._diag_variables

    @property
    def variable_groups(self) -> NamedItemList[VariableGroup]:
        return self._variable_groups

    #######
    # </properties subject to value inheritance>
    #######

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "EcuVariant":
        ecu_variant_raw = EcuVariantRaw.from_et(et_element, context)

        return EcuVariant(diag_layer_raw=ecu_variant_raw)

    def __post_init__(self) -> None:
        super().__post_init__()

        odxassert(
            isinstance(self.diag_layer_raw, EcuVariantRaw),
            "The raw diagnostic layer passed to EcuVariant "
            "must be a EcuVariantRaw")

    def __deepcopy__(self, memo: dict[int, Any]) -> Any:
        """Create a deep copy of the ECU variant

        Note that the copied diagnostic layer is not fully
        initialized, so `_finalize_init()` should to be called on it
        before it can be used normally.
        """

        result = super().__deepcopy__(memo)

        # note that the self.ecu_variant_raw object is *not* copied at
        # this place because the attribute points to the same object
        # as self.diag_layer_raw.
        result.ecu_variant_raw = deepcopy(self.ecu_variant_raw, memo)

        return result

    @override
    def _compute_value_inheritance(self, odxlinks: OdxLinkDatabase) -> None:
        super()._compute_value_inheritance(odxlinks)

        self._diag_variables = NamedItemList(self._compute_available_diag_variables(odxlinks))
        self._variable_groups = NamedItemList(self._compute_available_variable_groups(odxlinks))

    def _compute_available_diag_variables(self,
                                          odxlinks: OdxLinkDatabase) -> Iterable[DiagVariable]:

        def get_local_objects_fn(dl: DiagLayer) -> Iterable[DiagVariable]:
            if not isinstance(dl.diag_layer_raw, HasDiagVariables):
                return []

            return dl.diag_layer_raw.diag_variables

        def not_inherited_fn(parent_ref: ParentRef) -> list[str]:
            return parent_ref.not_inherited_variables

        return self._compute_available_objects(get_local_objects_fn, not_inherited_fn)

    def _compute_available_variable_groups(self,
                                           odxlinks: OdxLinkDatabase) -> Iterable[VariableGroup]:

        def get_local_objects_fn(dl: DiagLayer) -> Iterable[VariableGroup]:
            if not isinstance(dl.diag_layer_raw, HasVariableGroups):
                return []

            return dl.diag_layer_raw.variable_groups

        def not_inherited_fn(parent_ref: ParentRef) -> list[str]:
            return []

        return self._compute_available_objects(get_local_objects_fn, not_inherited_fn)



# SPDX-License-Identifier: MIT
from xml.etree import ElementTree

import pytest
from packaging.version import Version

from odxtools.basevariantpattern import BaseVariantPattern
from odxtools.ecuvariantpattern import EcuVariantPattern
from odxtools.exceptions import OdxError
from odxtools.odxdoccontext import OdxDocContext
from odxtools.odxlink import DocType, OdxDocFragment

doc_frags = [OdxDocFragment(doc_name="pytest", doc_type=DocType.CONTAINER)]


@pytest.fixture()
def valid_evp_et() -> ElementTree.Element:
    xml = """
    <ECU-VARIANT-PATTERNS>
        <ECU-VARIANT-PATTERN>
            <MATCHING-PARAMETERS>
                <MATCHING-PARAMETER>
                    <EXPECTED-VALUE>65535</EXPECTED-VALUE>
                    <DIAG-COMM-SNREF SHORT-NAME="IdentService"/>
                    <OUT-PARAM-IF-SNREF SHORT-NAME="Id"/>
                </MATCHING-PARAMETER>
            </MATCHING-PARAMETERS>
        </ECU-VARIANT-PATTERN>
        <ECU-VARIANT-PATTERN>
            <MATCHING-PARAMETERS>
                <MATCHING-PARAMETER>
                    <EXPECTED-VALUE>255</EXPECTED-VALUE>
                    <DIAG-COMM-SNREF SHORT-NAME="SerialNumber"/>
                    <OUT-PARAM-IF-SNREF SHORT-NAME="Value"/>
                </MATCHING-PARAMETER>
                <MATCHING-PARAMETER>
                    <EXPECTED-VALUE>supplier_A</EXPECTED-VALUE>
                    <DIAG-COMM-SNREF SHORT-NAME="SupplierInfo"/>
                    <OUT-PARAM-IF-SNPATHREF SHORT-NAME-PATH="name.english"/>
                </MATCHING-PARAMETER>
            </MATCHING-PARAMETERS>
        </ECU-VARIANT-PATTERN>
    </ECU-VARIANT-PATTERNS>
    """
    return ElementTree.fromstring(xml)


@pytest.fixture()
def valid_bvp_et() -> ElementTree.Element:
    xml = """
    <BASE-VARIANT-PATTERNS>
        <BASE-VARIANT-PATTERN>
            <MATCHING-BASE-VARIANT-PARAMETERS>
                <MATCHING-BASE-VARIANT-PARAMETER>
                    <EXPECTED-VALUE>1234</EXPECTED-VALUE>
                    <DIAG-COMM-SNREF SHORT-NAME="Identify"/>
                    <OUT-PARAM-IF-SNREF SHORT-NAME="ECU_ID"/>
                </MATCHING-BASE-VARIANT-PARAMETER>
            </MATCHING-BASE-VARIANT-PARAMETERS>
        </BASE-VARIANT-PATTERN>
        <BASE-VARIANT-PATTERN>
            <MATCHING-BASE-VARIANT-PARAMETERS>
                <MATCHING-BASE-VARIANT-PARAMETER>
                    <EXPECTED-VALUE>234</EXPECTED-VALUE>
                    <USE-PHYSICAL-ADDRESSING>true</USE-PHYSICAL-ADDRESSING>
                    <DIAG-COMM-SNREF SHORT-NAME="SerialInfo"/>
                    <OUT-PARAM-IF-SNREF SHORT-NAME="Val"/>
                </MATCHING-BASE-VARIANT-PARAMETER>
                <MATCHING-BASE-VARIANT-PARAMETER>
                    <EXPECTED-VALUE>OEM</EXPECTED-VALUE>
                    <USE-PHYSICAL-ADDRESSING>false</USE-PHYSICAL-ADDRESSING>
                    <DIAG-COMM-SNREF SHORT-NAME="ManufacturerInfo"/>
                    <OUT-PARAM-IF-SNPATHREF SHORT-NAME-PATH="info.type"/>
                </MATCHING-BASE-VARIANT-PARAMETER>
            </MATCHING-BASE-VARIANT-PARAMETERS>
        </BASE-VARIANT-PATTERN>
    </BASE-VARIANT-PATTERNS>
    """
    return ElementTree.fromstring(xml)


@pytest.fixture()
def invalid_evp_et() -> ElementTree.Element:
    xml = """
    <ECU-VARIANT-PATTERNS>
        <ECU-VARIANT-PATTERN>
            <MATCHING-PARAMETERS>
            </MATCHING-PARAMETERS>
        </ECU-VARIANT-PATTERN>
    </ECU-VARIANT-PATTERNS>
    """
    return ElementTree.fromstring(xml)


def test_create_bvp_from_et(valid_bvp_et: ElementTree.Element) -> None:
    base_variant_patterns = [
        BaseVariantPattern.from_et(elem, OdxDocContext(Version("2.2.0"), doc_frags))
        for elem in valid_bvp_et.iterfind("BASE-VARIANT-PATTERN")
    ]
    assert len(base_variant_patterns) == 2
    assert base_variant_patterns[0].matching_base_variant_parameters[0].use_physical_addressing
    assert base_variant_patterns[0].matching_base_variant_parameters[0].matches({"ECU_ID": 1234})
    assert base_variant_patterns[1].matching_base_variant_parameters[0].use_physical_addressing
    assert base_variant_patterns[1].matching_base_variant_parameters[0].matches({"Val": 234})
    assert not base_variant_patterns[1].matching_base_variant_parameters[1].use_physical_addressing
    assert base_variant_patterns[1].matching_base_variant_parameters[1].matches(
        {"info": {
            "type": "OEM"
        }})

    assert not base_variant_patterns[1].matching_base_variant_parameters[1].matches(
        {"info": {
            "type": 123
        }})

    assert not base_variant_patterns[1].matching_base_variant_parameters[1].matches(
        {"info": {
            "type": "scammer"
        }})

    assert not base_variant_patterns[1].matching_base_variant_parameters[1].matches(
        {"nice2know": {
            "type": "OEM"
        }})


def test_create_evp_from_et(valid_evp_et: ElementTree.Element) -> None:
    ecu_variant_patterns = [
        EcuVariantPattern.from_et(elem, OdxDocContext(Version("2.2.0"), doc_frags))
        for elem in valid_evp_et.iterfind("ECU-VARIANT-PATTERN")
    ]
    assert len(ecu_variant_patterns) == 2
    assert ecu_variant_patterns[0].matching_parameters[0].matches({"Id": 0xFFFF})
    assert ecu_variant_patterns[1].matching_parameters[0].matches({"Value": 0xFF})
    assert ecu_variant_patterns[1].matching_parameters[1].matches(
        {"name": {
            "english": "supplier_A"
        }})


def test_create_invalid_evp_from_et(invalid_evp_et: ElementTree.Element) -> None:
    with pytest.raises(OdxError):
        for x in invalid_evp_et.iterfind("ECU-VARIANT-PATTERN"):
            EcuVariantPattern.from_et(x, OdxDocContext(Version("2.2.0"), doc_frags))



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from .encoding import Encoding, get_string_encoding
from .exceptions import DecodeError, odxassert, odxraise, strict_mode
from .odxtypes import AtomicOdxType, DataType, ParameterValue

try:
    import bitstruct.c as bitstruct
except ImportError:
    import bitstruct

if TYPE_CHECKING:
    from .parameters.parameter import Parameter
    from .tablerow import TableRow


@dataclass
class DecodeState:
    """Utility class to be used while decoding a message."""

    #: bytes to be decoded
    coded_message: bytes

    #: Absolute position of the origin
    #:
    #: i.e., the absolute byte position to which all relative positions
    #: refer to, e.g. the position of the first byte of a structure.
    origin_byte_position: int = 0

    #: Absolute position of the next undecoded byte to be considered
    #:
    #: (if not explicitly specified by the object to be decoded.)
    cursor_byte_position: int = 0

    #: the bit position [0, 7] where the object to be extracted begins
    #:
    #: If bit position is undefined (`None`), the object to be extracted
    #: starts at bit 0.
    cursor_bit_position: int = 0

    #: values of the length key parameters decoded so far
    length_keys: dict[str, int] = field(default_factory=dict)

    #: values of the table key parameters decoded so far
    table_keys: dict[str, "TableRow"] = field(default_factory=dict)

    #: List of parameters that have been decoded so far. The journal
    #: is used by some types of parameters which depend on the values of
    #: other parameters; i.e., environment data description parameters
    journal: list[tuple["Parameter", ParameterValue | None]] = field(default_factory=list)

    def extract_atomic_value(
        self,
        *,
        bit_length: int,
        base_data_type: DataType,
        base_type_encoding: Encoding | None,
        is_highlow_byte_order: bool,
    ) -> AtomicOdxType:
        """Extract an internal value from a blob of raw bytes.

        :return: Tuple with the internal value of the object and the
                 byte position of the first undecoded byte after the
                 extracted object.
        """
        # If the bit length is zero, return "empty" values of each type
        if bit_length == 0:
            return base_data_type.python_type()

        if base_data_type == DataType.A_FLOAT32 and bit_length != 32:
            odxraise("The bit length of FLOAT32 values must be 32 bits")
            bit_length = 32
        elif base_data_type == DataType.A_FLOAT64 and bit_length != 64:
            odxraise("The bit length of FLOAT64 values must be 64 bits")
            bit_length = 64

        byte_length = (bit_length + self.cursor_bit_position + 7) // 8
        if self.cursor_byte_position + byte_length > len(self.coded_message):
            raise DecodeError(f"Expected a longer message.")
        extracted_bytes = self.coded_message[self.cursor_byte_position:self.cursor_byte_position +
                                             byte_length]

        # Apply byteorder for numerical objects. Note that doing this
        # here might lead to garbage data being included in the result
        # if the data to be extracted is not byte aligned and crosses
        # byte boundaries, but it is what the specification says.
        if not is_highlow_byte_order and base_data_type in [
                DataType.A_INT32,
                DataType.A_UINT32,
                DataType.A_FLOAT32,
                DataType.A_FLOAT64,
        ]:
            extracted_bytes = extracted_bytes[::-1]

        padding = (8 - (bit_length + self.cursor_bit_position) % 8) % 8
        raw_value, = bitstruct.unpack_from(
            f"{base_data_type.bitstruct_format_letter}{bit_length}",
            extracted_bytes,
            offset=padding)
        internal_value: AtomicOdxType

        # Deal with raw byte fields, ...
        if base_data_type == DataType.A_BYTEFIELD:
            odxassert(
                base_type_encoding in (None, Encoding.NONE, Encoding.BCD_P, Encoding.BCD_UP),
                f"Illegal encoding '{base_type_encoding}' for A_BYTEFIELD")

            # note that we do not ensure that BCD-encoded byte fields
            # only represent "legal" values
            internal_value = raw_value

        # ... string types, ...
        elif base_data_type in (DataType.A_UTF8STRING, DataType.A_ASCIISTRING,
                                DataType.A_UNICODE2STRING):
            text_errors = 'strict' if strict_mode else 'replace'
            str_encoding = get_string_encoding(base_data_type, base_type_encoding,
                                               is_highlow_byte_order)
            if str_encoding is not None:
                internal_value = raw_value.decode(str_encoding, errors=text_errors)
            else:
                internal_value = "ERROR"

        # ... signed integers, ...
        elif base_data_type == DataType.A_INT32:
            if base_type_encoding == Encoding.ONEC:
                # one-complement
                sign_bit = 1 << (bit_length - 1)
                if raw_value < sign_bit:
                    internal_value = raw_value
                else:
                    # python defines the bitwise inversion of a
                    # positive integer value x as ~x = -(x + 1).
                    internal_value = -((1 << bit_length) - raw_value - 1)
            elif base_type_encoding in (None, Encoding.TWOC):
                # two-complement
                sign_bit = 1 << (bit_length - 1)
                if raw_value < sign_bit:
                    internal_value = raw_value
                else:
                    internal_value = -((1 << bit_length) - raw_value)
            elif base_type_encoding == Encoding.SM:
                # sign-magnitude
                sign_bit = 1 << (bit_length - 1)
                if raw_value < sign_bit:
                    internal_value = raw_value
                else:
                    internal_value = -(raw_value - sign_bit)
            else:
                odxraise(f"Illegal encoding ({base_type_encoding}) specified for "
                         f"{base_data_type.value}")

                if base_type_encoding == Encoding.BCD_P:
                    internal_value = self.__decode_bcd_p(raw_value)
                elif base_type_encoding == Encoding.BCD_UP:
                    internal_value = self.__decode_bcd_up(raw_value)
                else:
                    internal_value = raw_value

        # ... unsigned integers, ...
        elif base_data_type == DataType.A_UINT32:
            if base_type_encoding == Encoding.BCD_P:
                internal_value = self.__decode_bcd_p(raw_value)
            elif base_type_encoding == Encoding.BCD_UP:
                internal_value = self.__decode_bcd_up(raw_value)
            elif base_type_encoding in (None, Encoding.NONE):
                # no encoding
                internal_value = raw_value
            else:
                odxraise(f"Illegal encoding ({base_type_encoding}) specified for "
                         f"{base_data_type.value}")

                internal_value = raw_value

        # ... and others (floating point values)
        else:
            odxassert(base_data_type in (DataType.A_FLOAT32, DataType.A_FLOAT64))
            odxassert(
                base_type_encoding in (None, Encoding.NONE),
                f"Specified illegal encoding '{base_type_encoding}' for float object")

            internal_value = float(raw_value)

        self.cursor_byte_position += byte_length
        self.cursor_bit_position = 0

        return internal_value

    @staticmethod
    def __decode_bcd_p(value: int) -> int:
        # packed BCD
        result = 0
        factor = 1
        while value > 0:
            result += (value & 0xf) * factor
            factor *= 10
            value >>= 4

        return result

    @staticmethod
    def __decode_bcd_up(value: int) -> int:
        # unpacked BCD
        result = 0
        factor = 1
        while value > 0:
            result += (value & 0xf) * factor
            factor *= 10
            value >>= 8

        return result



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .description import Description
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext
from .xdoc import XDoc


@dataclass(kw_only=True)
class RelatedDoc:
    xdoc: XDoc | None = None
    description: Description | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "RelatedDoc":
        xdoc: XDoc | None = None
        if (xdoc_elem := et_element.find("XDOC")) is not None:
            xdoc = XDoc.from_et(xdoc_elem, context)
        description = Description.from_et(et_element.find("DESC"), context)

        return RelatedDoc(
            xdoc=xdoc,
            description=description,
        )

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {}

        if self.xdoc:
            result.update(self.xdoc._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if self.xdoc:
            self.xdoc._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        if self.xdoc:
            self.xdoc._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .basecomparam import BaseComparam
from .dataobjectproperty import DataObjectProperty
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .odxtypes import AtomicOdxType
from .snrefcontext import SnRefContext
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class Comparam(BaseComparam):
    physical_default_value_raw: str
    dop_ref: OdxLinkRef

    @property
    def physical_default_value(self) -> AtomicOdxType:
        return self._physical_default_value

    @property
    def dop(self) -> DataObjectProperty:
        """The data object property applicable to this parameter."""
        return self._dop

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "Comparam":
        kwargs = dataclass_fields_asdict(BaseComparam.from_et(et_element, context))

        physical_default_value_raw = odxrequire(et_element.findtext("PHYSICAL-DEFAULT-VALUE"))
        dop_ref = odxrequire(OdxLinkRef.from_et(et_element.find("DATA-OBJECT-PROP-REF"), context))

        return Comparam(
            dop_ref=dop_ref, physical_default_value_raw=physical_default_value_raw, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return super()._build_odxlinks()

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        """Resolves the reference to the dop"""
        super()._resolve_odxlinks(odxlinks)

        self._dop = odxlinks.resolve(self.dop_ref, DataObjectProperty)
        self._physical_default_value = self._dop.physical_type.base_data_type.from_string(
            self.physical_default_value_raw)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
import datetime
import inspect
import mimetypes
import os
import time
import zipfile
from typing import Any

import jinja2

import odxtools_porsche

from .database import Database
from .odxtypes import bool_to_odxstr

odxdatabase: Database | None = None


def jinja2_odxraise_helper(msg: str) -> None:
    raise Exception(msg)


def get_parent_container_name(dl_short_name: str) -> str:
    """
    Given the short name of a diagnostic layer, return the name of a container
    which the layer is part of.

    If no such container exists, a `RuntimeException` is thrown.
    """

    assert odxdatabase is not None

    for dlc in odxdatabase.diag_layer_containers:
        if dl_short_name in [dl.short_name for dl in dlc.diag_layers]:
            return dlc.short_name

    raise RuntimeError(f"get_parent_container_name() could not determine a "
                       f"container for diagnostic layer '{dl_short_name}'.")


def make_xml_attrib(attrib_name: str, attrib_val: Any | None) -> str:
    if attrib_val is None:
        return ""

    return f' {attrib_name}="{attrib_val}"'


def make_bool_xml_attrib(attrib_name: str, attrib_val: bool | None) -> str:
    if attrib_val is None:
        return ""

    return make_xml_attrib(attrib_name, bool_to_odxstr(attrib_val))


__module_filename = inspect.getsourcefile(odxtools_porsche)
assert isinstance(__module_filename, str)
__templates_dir = os.path.sep.join([os.path.dirname(__module_filename), "templates"])


def write_pdx_file(
    output_file_name: str,
    database: Database,
    templates_dir: str = __templates_dir,
) -> bool:
    """
    Write an internalized database to a PDX file.
    """
    global odxdatabase

    odxdatabase = database

    file_index = []
    with zipfile.ZipFile(output_file_name, mode="w", compression=zipfile.ZIP_DEFLATED) as zf:

        # write all files in the templates directory
        for root, _, files in os.walk(templates_dir):
            for template_file_name in files:
                # we are not interested in the autosave garbage of
                # editors...
                if template_file_name.startswith("."):
                    continue
                elif template_file_name.startswith("#"):
                    continue
                elif template_file_name.endswith("~"):
                    continue
                elif template_file_name.endswith(".bak"):
                    continue
                elif template_file_name.endswith(".xml.jinja2"):
                    continue
                elif template_file_name.endswith(".odx-cs"):
                    # we don't copy the comparam subset files (they
                    # are written based on the database)
                    continue

                template_file_mime_type = None
                if template_file_name.endswith(".odx-cs"):
                    template_file_mime_type = "application/x-asam.odx.odx-cs"
                elif template_file_name.endswith(".odx-d"):
                    template_file_mime_type = "application/x-asam.odx.odx-d"

                guessed_mime_type, guessed_encoding = mimetypes.guess_type(template_file_name)
                if template_file_mime_type is None and guessed_mime_type is not None:
                    template_file_mime_type = guessed_mime_type
                else:
                    template_file_mime_type = "application/octet-stream"

                in_path = [root]
                in_path.append(template_file_name)
                in_file_name = os.path.sep.join(in_path)

                template_file_stats = os.stat(in_file_name)
                template_file_cdate = datetime.datetime.fromtimestamp(template_file_stats.st_ctime)
                template_file_creation_date = template_file_cdate.strftime("%Y-%m-%dT%H:%M:%S")
                file_index.append(
                    (template_file_name, template_file_creation_date, template_file_mime_type))
                with zf.open(template_file_name, "w") as out_file:
                    out_file.write(open(in_file_name, "rb").read())

        # write the auxiliary files
        for output_file_name, data_file in database.auxiliary_files.items():
            file_cdate = datetime.datetime.fromtimestamp(time.time())
            creation_date = file_cdate.strftime("%Y-%m-%dT%H:%M:%S")

            mime_type = None
            if output_file_name.endswith(".odx-cs"):
                mime_type = "application/x-asam.odx.odx-cs"
            elif output_file_name.endswith(".odx-d"):
                mime_type = "application/x-asam.odx.odx-d"

            guessed_mime_type, guessed_encoding = mimetypes.guess_type(output_file_name)
            if mime_type is None and guessed_mime_type is not None:
                mime_type = guessed_mime_type
            else:
                mime_type = "application/octet-stream"

            zf_name = os.path.basename(output_file_name)
            with zf.open(zf_name, "w") as out_file:
                file_index.append((zf_name, creation_date, mime_type))
                out_file.write(data_file.read())

        jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(templates_dir))
        jinja_env.globals["getattr"] = getattr
        jinja_env.globals["hasattr"] = hasattr
        jinja_env.globals["odxraise"] = jinja2_odxraise_helper
        jinja_env.globals["make_xml_attrib"] = make_xml_attrib
        jinja_env.globals["make_bool_xml_attrib"] = make_bool_xml_attrib
        jinja_env.globals["get_parent_container_name"] = get_parent_container_name

        vars: dict[str, Any] = {}
        vars["odxtools_version"] = odxtools.__version__
        vars["database"] = database

        # write the communication parameter subsets
        comparam_subset_tpl = jinja_env.get_template("comparam-subset.odx-cs.xml.jinja2")
        for comparam_subset in database.comparam_subsets:
            zf_file_name = f"{comparam_subset.short_name}.odx-cs"
            zf_file_cdate = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
            zf_mime_type = "application/x-asam.odx.odx-cs"

            vars["comparam_subset"] = comparam_subset

            file_index.append((zf_file_name, zf_file_cdate, zf_mime_type))

            zf.writestr(zf_file_name, comparam_subset_tpl.render(**vars))

            del vars["comparam_subset"]

        # write the communication parameter specs
        comparam_spec_tpl = jinja_env.get_template("comparam-spec.odx-c.xml.jinja2")
        for comparam_spec in database.comparam_specs:
            zf_file_name = f"{comparam_spec.short_name}.odx-c"
            zf_file_cdate = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
            zf_mime_type = "application/x-asam.odx.odx-c"

            vars["comparam_spec"] = comparam_spec

            file_index.append((zf_file_name, zf_file_cdate, zf_mime_type))

            zf.writestr(zf_file_name, comparam_spec_tpl.render(**vars))

            del vars["comparam_spec"]

        # write the actual diagnostic data.
        dlc_tpl = jinja_env.get_template("diag_layer_container.odx-d.xml.jinja2")
        for dlc in database.diag_layer_containers:
            vars["dlc"] = dlc

            file_name = f"{dlc.short_name}.odx-d"
            file_cdate = datetime.datetime.now()
            creation_date = file_cdate.strftime("%Y-%m-%dT%H:%M:%S")
            mime_type = "application/x-asam.odx.odx-d"

            file_index.append((file_name, creation_date, mime_type))
            zf.writestr(file_name, dlc_tpl.render(**vars))
            del vars["dlc"]

        # write the index.xml file
        vars["file_index"] = file_index
        index_tpl = jinja_env.get_template("index.xml.jinja2")
        text = index_tpl.render(**vars)
        zf.writestr("index.xml", text)

    return True



from dataclasses import dataclass
from xml.etree import ElementTree

from .description import Description
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkId
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class NamedElement:
    short_name: str
    long_name: str | None = None
    description: Description | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "NamedElement":

        return NamedElement(
            short_name=odxrequire(et_element.findtext("SHORT-NAME")),
            long_name=et_element.findtext("LONG-NAME"),
            description=Description.from_et(et_element.find("DESC"), context),
        )


@dataclass(kw_only=True)
class IdentifiableElement(NamedElement):
    odx_id: OdxLinkId
    oid: str | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "IdentifiableElement":

        kwargs = dataclass_fields_asdict(NamedElement.from_et(et_element, context))

        odx_id = odxrequire(OdxLinkId.from_et(et_element, context))
        oid = et_element.get("OID")

        return IdentifiableElement(**kwargs, odx_id=odx_id, oid=oid)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .companyspecificinfo import CompanySpecificInfo
from .element import IdentifiableElement
from .exceptions import odxrequire
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext
from .teammember import TeamMember
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class CompanyData(IdentifiableElement):
    roles: list[str] = field(default_factory=list)
    team_members: NamedItemList[TeamMember] = field(default_factory=NamedItemList)
    company_specific_info: CompanySpecificInfo | None = None

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "CompanyData":

        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        roles = []
        if (roles_elem := et_element.find("ROLES")) is not None:
            roles = [odxrequire(role.text) for role in roles_elem.iterfind("ROLE")]
        team_members = [
            TeamMember.from_et(tm, context)
            for tm in et_element.iterfind("TEAM-MEMBERS/TEAM-MEMBER")
        ]
        company_specific_info = None
        if (company_specific_info_elem := et_element.find("COMPANY-SPECIFIC-INFO")) is not None:
            company_specific_info = CompanySpecificInfo.from_et(company_specific_info_elem, context)

        return CompanyData(
            roles=roles,
            team_members=NamedItemList(team_members),
            company_specific_info=company_specific_info,
            **kwargs,
        )

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {self.odx_id: self}

        for tm in self.team_members:
            result.update(tm._build_odxlinks())

        if self.company_specific_info:
            result.update(self.company_specific_info._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        for tm in self.team_members:
            tm._resolve_odxlinks(odxlinks)

        if self.company_specific_info:
            self.company_specific_info._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        for tm in self.team_members:
            tm._resolve_snrefs(context)

        if self.company_specific_info:
            self.company_specific_info._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from xml.etree import ElementTree

from typing_extensions import override

from .exceptions import odxassert
from .matchingbasevariantparameter import MatchingBaseVariantParameter
from .matchingparameter import MatchingParameter
from .odxdoccontext import OdxDocContext
from .variantpattern import VariantPattern


@dataclass(kw_only=True)
class BaseVariantPattern(VariantPattern):
    """Base variant patterns are variant patterns used to identify the
    base variant of an ECU.
    """
    matching_base_variant_parameters: list[MatchingBaseVariantParameter] = field(
        default_factory=list)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "BaseVariantPattern":

        matching_base_variant_parameters = [
            MatchingBaseVariantParameter.from_et(mbvp_el, context)
            for mbvp_el in et_element.iterfind("MATCHING-BASE-VARIANT-PARAMETERS/"
                                               "MATCHING-BASE-VARIANT-PARAMETER")
        ]

        odxassert(len(matching_base_variant_parameters) > 0)  # required by ISO 22901-1 Figure 141
        return BaseVariantPattern(matching_base_variant_parameters=matching_base_variant_parameters)

    @override
    def get_matching_parameters(self
                               ) -> list[MatchingParameter] | list[MatchingBaseVariantParameter]:
        return self.matching_base_variant_parameters



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .compumethods.limit import Limit
from .element import NamedElement
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from .odxtypes import AtomicOdxType, DataType
from .snrefcontext import SnRefContext
from .structure import Structure
from .utils import dataclass_fields_asdict


@dataclass(kw_only=True)
class MultiplexerCase(NamedElement):
    """This class represents a case which represents a range of keys of a multiplexer."""

    structure_ref: OdxLinkRef | None = None
    structure_snref: str | None = None
    lower_limit: Limit
    upper_limit: Limit

    @property
    def structure(self) -> Structure | None:
        return self._structure

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "MultiplexerCase":
        """Reads a case for a Multiplexer."""
        kwargs = dataclass_fields_asdict(NamedElement.from_et(et_element, context))
        structure_ref = OdxLinkRef.from_et(et_element.find("STRUCTURE-REF"), context)
        structure_snref = None
        if (structure_snref_elem := et_element.find("STRUCTURE-SNREF")) is not None:
            structure_snref = odxrequire(structure_snref_elem.get("SHORT-NAME"))

        lower_limit = Limit.limit_from_et(
            odxrequire(et_element.find("LOWER-LIMIT")),
            context,
            value_type=None,
        )
        upper_limit = Limit.limit_from_et(
            odxrequire(et_element.find("UPPER-LIMIT")),
            context,
            value_type=None,
        )

        return MultiplexerCase(
            structure_ref=structure_ref,
            structure_snref=structure_snref,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        raise RuntimeError("Calling MultiplexerCase._resolve_odxlinks() is not allowed. "
                           "Use ._mux_case_resolve_odxlinks().")

    def _mux_case_resolve_odxlinks(self, odxlinks: OdxLinkDatabase, *,
                                   key_physical_type: DataType) -> None:
        self._structure = None
        if self.structure_ref:
            self._structure = odxlinks.resolve(self.structure_ref, Structure)

        self.lower_limit.set_value_type(key_physical_type)
        self.upper_limit.set_value_type(key_physical_type)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        if self.structure_snref:
            ddds = odxrequire(context.diag_layer).diag_data_dictionary_spec
            self._structure = resolve_snref(self.structure_snref, ddds.structures, Structure)

    def applies(self, value: AtomicOdxType) -> bool:
        return self.lower_limit.complies_to_lower(value) \
            and self.upper_limit.complies_to_upper(value)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from .compumethods.limit import Limit
from .description import Description
from .exceptions import odxraise, odxrequire
from .odxdoccontext import OdxDocContext
from .odxtypes import DataType
from .validtype import ValidType


@dataclass(kw_only=True)
class ScaleConstr:
    """This class represents a SCALE-CONSTR.
    """

    short_label: str | None = None
    description: Description | None = None
    lower_limit: Limit
    upper_limit: Limit
    validity: ValidType
    value_type: DataType

    @staticmethod
    def scale_constr_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                             value_type: DataType) -> "ScaleConstr":
        short_label = et_element.findtext("SHORT-LABEL")
        description = Description.from_et(et_element.find("DESC"), context)
        lower_limit = Limit.limit_from_et(
            odxrequire(et_element.find("LOWER-LIMIT")), context, value_type=value_type)
        upper_limit = Limit.limit_from_et(
            odxrequire(et_element.find("UPPER-LIMIT")), context, value_type=value_type)

        validity_str = odxrequire(et_element.get("VALIDITY"))
        try:
            validity = ValidType(validity_str)
        except ValueError:
            odxraise(f"Encountered unknown Validity '{validity_str}'")

        return ScaleConstr(
            short_label=short_label,
            description=description,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            validity=validity,
            value_type=value_type)



# SPDX-License-Identifier: MIT
import getpass
from dataclasses import dataclass
from datetime import datetime
from xml.etree import ElementTree

from typing_extensions import override

from ..encodestate import EncodeState
from ..exceptions import odxraise, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxtypes import ParameterValue
from ..utils import dataclass_fields_asdict
from .parameter import ParameterType
from .parameterwithdop import ParameterWithDOP

# The SYSTEM parameter types mandated by the ODX 2.2 standard. Users
# are free to specify additional types, but these must be handled
# (cf. table 5 in section 7.3.5.4 of the ASAM ODX 2.2 specification
# document.)
PREDEFINED_SYSPARAM_VALUES = [
    "TIMESTAMP", "SECOND", "MINUTE", "HOUR", "TIMEZONE", "DAY", "WEEK", "MONTH", "YEAR", "CENTURY",
    "TESTERID", "USERID"
]


@dataclass(kw_only=True)
class SystemParameter(ParameterWithDOP):
    sysparam: str

    @property
    @override
    def parameter_type(self) -> ParameterType:
        return "SYSTEM"

    @property
    @override
    def is_required(self) -> bool:
        # if a SYSTEM parameter is not specified explicitly, its value
        # can be determined from the operating system if it is type is
        # predefined
        return self.sysparam not in PREDEFINED_SYSPARAM_VALUES

    @property
    @override
    def is_settable(self) -> bool:
        return True

    @staticmethod
    @override
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "SystemParameter":
        kwargs = dataclass_fields_asdict(ParameterWithDOP.from_et(et_element, context))

        sysparam = odxrequire(et_element.get("SYSPARAM"))

        return SystemParameter(sysparam=sysparam, **kwargs)

    @override
    def _encode_positioned_into_pdu(self, physical_value: ParameterValue | None,
                                    encode_state: EncodeState) -> None:
        if physical_value is None:
            # determine the value to be encoded automatically
            now = datetime.now()
            if self.sysparam == "TIMESTAMP":
                physical_value = round(now.timestamp() * 1000).to_bytes(8, "big")
            elif self.sysparam == "SECOND":
                physical_value = now.second
            elif self.sysparam == "MINUTE":
                physical_value = now.minute
            elif self.sysparam == "HOUR":
                physical_value = now.hour
            elif self.sysparam == "TIMEZONE":
                if (utc_offset := now.astimezone().utcoffset()) is not None:
                    physical_value = utc_offset.seconds // 60
                else:
                    physical_value = 0
            elif self.sysparam == "DAY":
                physical_value = now.day
            elif self.sysparam == "WEEK":
                physical_value = now.isocalendar()[1]
            elif self.sysparam == "MONTH":
                physical_value = now.month
            elif self.sysparam == "YEAR":
                physical_value = now.year
            elif self.sysparam == "CENTURY":
                physical_value = now.year // 100
            elif self.sysparam == "TESTERID":
                physical_value = "odxtools".encode("latin1")
            elif self.sysparam == "USERID":
                physical_value = getpass.getuser().encode("latin1")
            else:
                odxraise(f"Unknown system parameter type '{self.sysparam}'")
                physical_value = 0

        self.dop.encode_into_pdu(physical_value, encode_state=encode_state)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .database import Database
    from .diaglayers.diaglayer import DiagLayer
    from .diagservice import DiagService
    from .parameters.parameter import Parameter
    from .request import Request
    from .response import Response
    from .singleecujob import SingleEcuJob
    from .statechart import StateChart


@dataclass(kw_only=True)
class SnRefContext:
    """Represents the context for which a short name reference ought
    to be resolved
    """

    database: Optional["Database"] = None
    diag_layer: Optional["DiagLayer"] = None
    diag_service: Optional["DiagService"] = None
    single_ecu_job: Optional["SingleEcuJob"] = None
    request: Optional["Request"] = None
    response: Optional["Response"] = None
    parameters: list["Parameter"] | None = None
    state_chart: Optional["StateChart"] = None



# SPDX-License-Identifier: MIT
import inspect
import os
import unittest
from io import BytesIO
from typing import NamedTuple, cast
from xml.etree import ElementTree

import jinja2
from packaging.version import Version

import odxtools_porsche
from odxtools.additionalaudience import AdditionalAudience
from odxtools.audience import Audience
from odxtools.compumethods.compucategory import CompuCategory
from odxtools.compumethods.compuconst import CompuConst
from odxtools.compumethods.compuinternaltophys import CompuInternalToPhys
from odxtools.compumethods.compurationalcoeffs import CompuRationalCoeffs
from odxtools.compumethods.compuscale import CompuScale
from odxtools.compumethods.limit import Limit
from odxtools.compumethods.linearcompumethod import LinearCompuMethod
from odxtools.compumethods.texttablecompumethod import TexttableCompuMethod
from odxtools.database import Database
from odxtools.dataobjectproperty import DataObjectProperty
from odxtools.description import Description
from odxtools.diaglayers.diaglayertype import DiagLayerType
from odxtools.diaglayers.ecuvariant import EcuVariant
from odxtools.diaglayers.ecuvariantraw import EcuVariantRaw
from odxtools.exceptions import odxrequire
from odxtools.functionalclass import FunctionalClass
from odxtools.inputparam import InputParam
from odxtools.library import Library
from odxtools.nameditemlist import NamedItemList
from odxtools.negoutputparam import NegOutputParam
from odxtools.odxdoccontext import OdxDocContext
from odxtools.odxlink import DocType, OdxDocFragment, OdxLinkDatabase, OdxLinkId, OdxLinkRef
from odxtools.odxtypes import DataType
from odxtools.outputparam import OutputParam
from odxtools.physicaltype import PhysicalType
from odxtools.progcode import ProgCode
from odxtools.singleecujob import SingleEcuJob
from odxtools.standardlengthtype import StandardLengthType
from odxtools.writepdxfile import jinja2_odxraise_helper, make_bool_xml_attrib, make_xml_attrib

doc_frags = [OdxDocFragment("UnitTest", DocType.CONTAINER)]


class TestSingleEcuJob(unittest.TestCase):

    def setUp(self) -> None:
        """Create three objects:

        * self.singleecujob_object: SingleEcuJob - job to be tested
        * self.context: NamedTuple - elements referenced by the SingleEcuJob
        * self.singleecujob_odx: string - odx description of self.singleecujob_object
        """
        super().setUp()

        class Context(NamedTuple):
            """odx elements referenced by the tested single ECU job, i.e., elements needed in the `odxlinks` when resolving references"""

            extensiveTask: FunctionalClass
            specialAudience: AdditionalAudience
            inputDOP: DataObjectProperty
            outputDOP: DataObjectProperty
            negOutputDOP: DataObjectProperty

        self.context = Context(
            extensiveTask=FunctionalClass(
                odx_id=OdxLinkId("ID.extensiveTask", doc_frags),
                short_name="extensiveTask",
            ),
            specialAudience=AdditionalAudience(
                odx_id=OdxLinkId("ID.specialAudience", doc_frags),
                short_name="specialAudience",
            ),
            inputDOP=DataObjectProperty(
                odx_id=OdxLinkId("ID.inputDOP", doc_frags),
                short_name="inputDOP",
                diag_coded_type=StandardLengthType(
                    base_data_type=DataType.A_INT32,
                    bit_length=1,
                ),
                physical_type=PhysicalType(base_data_type=DataType.A_UNICODE2STRING),
                compu_method=TexttableCompuMethod(
                    category=CompuCategory.TEXTTABLE,
                    physical_type=DataType.A_UNICODE2STRING,
                    compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                        CompuScale(
                            short_label="yes",
                            lower_limit=Limit(value_raw="0", value_type=DataType.A_INT32),
                            compu_const=CompuConst(vt="Yes!", data_type=DataType.A_UTF8STRING),
                            domain_type=DataType.A_INT32,
                            range_type=DataType.A_UNICODE2STRING,
                        ),
                        CompuScale(
                            short_label="no",
                            lower_limit=Limit(value_raw="1", value_type=DataType.A_INT32),
                            compu_const=CompuConst(vt="Yes!", data_type=DataType.A_UTF8STRING),
                            domain_type=DataType.A_INT32,
                            range_type=DataType.A_UNICODE2STRING),
                    ]),
                    internal_type=DataType.A_UINT32,
                ),
            ),
            outputDOP=DataObjectProperty(
                odx_id=OdxLinkId("ID.outputDOP", doc_frags),
                short_name="outputDOP",
                diag_coded_type=StandardLengthType(
                    base_data_type=DataType.A_INT32,
                    bit_length=1,
                ),
                physical_type=PhysicalType(base_data_type=DataType.A_UNICODE2STRING),
                compu_method=LinearCompuMethod(
                    category=CompuCategory.LINEAR,
                    compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                        CompuScale(
                            compu_rational_coeffs=CompuRationalCoeffs(
                                value_type=DataType.A_INT32,
                                numerators=[1, -1],
                                denominators=[1],
                            ),
                            domain_type=DataType.A_INT32,
                            range_type=DataType.A_INT32),
                    ]),
                    internal_type=DataType.A_UINT32,
                    physical_type=DataType.A_UINT32),
            ),
            negOutputDOP=DataObjectProperty(
                odx_id=OdxLinkId("ID.negOutputDOP", doc_frags),
                short_name="negOutputDOP",
                diag_coded_type=StandardLengthType(
                    base_data_type=DataType.A_INT32,
                    bit_length=1,
                ),
                physical_type=PhysicalType(base_data_type=DataType.A_UNICODE2STRING),
                compu_method=LinearCompuMethod(
                    category=CompuCategory.LINEAR,
                    compu_internal_to_phys=CompuInternalToPhys(compu_scales=[
                        CompuScale(
                            compu_rational_coeffs=CompuRationalCoeffs(
                                value_type=DataType.A_INT32,
                                numerators=[1, -1],
                                denominators=[1],
                            ),
                            domain_type=DataType.A_INT32,
                            range_type=DataType.A_INT32),
                    ]),
                    internal_type=DataType.A_UINT32,
                    physical_type=DataType.A_UINT32,
                ),
            ),
        )

        input_params = NamedItemList([
            InputParam(
                short_name="inputParam",
                physical_default_value="Yes!",
                dop_base_ref=OdxLinkRef.from_id(self.context.inputDOP.odx_id),
            )
        ])
        output_params = NamedItemList([
            OutputParam(
                odx_id=OdxLinkId("ID.outputParam", doc_frags),
                semantic="DATA",
                short_name="outputParam",
                long_name="The Output Param",
                description=Description.from_string("<p>The one and only output of this job.</p>"),
                dop_base_ref=OdxLinkRef.from_id(self.context.outputDOP.odx_id),
            )
        ])
        neg_output_params = NamedItemList([
            NegOutputParam(
                short_name="NegativeOutputParam",
                description=Description.from_string("<p>The one and only output of this job.</p>"),
                dop_base_ref=OdxLinkRef.from_id(self.context.negOutputDOP.odx_id),
            )
        ])

        self.singleecujob_object = SingleEcuJob(
            odx_id=OdxLinkId("ID.JumpStart", doc_frags),
            short_name="JumpStart",
            functional_class_refs=[OdxLinkRef.from_id(self.context.extensiveTask.odx_id)],
            audience=Audience(
                enabled_audience_refs=[OdxLinkRef.from_id(self.context.specialAudience.odx_id)],
                is_manufacturing_raw=False,
            ),
            prog_codes=[
                ProgCode(
                    code_file="abc.jar",
                    encryption="RSA512",
                    syntax="JAR",
                    revision="0.12.34",
                    entrypoint="CalledClass",
                    library_refs=[OdxLinkRef("my.favourite.lib", doc_frags)],
                )
            ],
            input_params=input_params,
            output_params=output_params,
            neg_output_params=neg_output_params,
        )

        self.singleecujob_odx = f"""
            <SINGLE-ECU-JOB ID="{self.singleecujob_object.odx_id.local_id}">
                <SHORT-NAME>{self.singleecujob_object.short_name}</SHORT-NAME>
                <FUNCT-CLASS-REFS>
                    <FUNCT-CLASS-REF ID-REF="{self.singleecujob_object.functional_class_refs[0].ref_id}"/>
                </FUNCT-CLASS-REFS>
                <AUDIENCE IS-MANUFACTURING="false">
                    <ENABLED-AUDIENCE-REFS>
                        <ENABLED-AUDIENCE-REF ID-REF="{cast(Audience, self.singleecujob_object.audience).enabled_audience_refs[0].ref_id}"/>
                    </ENABLED-AUDIENCE-REFS>
                </AUDIENCE>
                <PROG-CODES>
                    <PROG-CODE>
                        <CODE-FILE>{self.singleecujob_object.prog_codes[0].code_file}</CODE-FILE>
                        <ENCRYPTION>{self.singleecujob_object.prog_codes[0].encryption}</ENCRYPTION>
                        <SYNTAX>{self.singleecujob_object.prog_codes[0].syntax}</SYNTAX>
                        <REVISION>{self.singleecujob_object.prog_codes[0].revision}</REVISION>
                        <ENTRYPOINT>{self.singleecujob_object.prog_codes[0].entrypoint}</ENTRYPOINT>
                        <LIBRARY-REFS>
                            <LIBRARY-REF ID-REF="{self.singleecujob_object.prog_codes[0].library_refs[0].ref_id}"/>
                        </LIBRARY-REFS>
                    </PROG-CODE>
                </PROG-CODES>
                <INPUT-PARAMS>
                    <INPUT-PARAM>
                        <SHORT-NAME>{input_params[0].short_name}</SHORT-NAME>
                        <PHYSICAL-DEFAULT-VALUE>{input_params[0].physical_default_value}</PHYSICAL-DEFAULT-VALUE>
                        <DOP-BASE-REF ID-REF="{input_params[0].dop_base_ref.ref_id}"/>
                    </INPUT-PARAM>
                </INPUT-PARAMS>
                <OUTPUT-PARAMS>
                    <OUTPUT-PARAM ID="{output_params[0].odx_id.local_id}" SEMANTIC="{output_params[0].semantic}">
                        <SHORT-NAME>{output_params[0].short_name}</SHORT-NAME>
                        <LONG-NAME>{output_params[0].long_name}</LONG-NAME>
                        <DESC>{output_params[0].description}</DESC>
                        <DOP-BASE-REF ID-REF="{output_params[0].dop_base_ref.ref_id}"/>
                    </OUTPUT-PARAM>
                </OUTPUT-PARAMS>
                <NEG-OUTPUT-PARAMS>
                    <NEG-OUTPUT-PARAM>
                        <SHORT-NAME>{neg_output_params[0].short_name}</SHORT-NAME>
                        <DESC>{neg_output_params[0].description}</DESC>
                        <DOP-BASE-REF ID-REF="{neg_output_params[0].dop_base_ref.ref_id}"/>
                    </NEG-OUTPUT-PARAM>
                </NEG-OUTPUT-PARAMS>
            </SINGLE-ECU-JOB>
        """

    def test_read_odx(self) -> None:
        expected = self.singleecujob_object
        sample_single_ecu_job_odx = self.singleecujob_odx
        et_element = ElementTree.fromstring(sample_single_ecu_job_odx)
        sej = SingleEcuJob.from_et(et_element, OdxDocContext(Version("2.2.0"), doc_frags))
        self.assertEqual(expected.prog_codes, sej.prog_codes)
        self.assertEqual(expected.output_params, sej.output_params)
        self.assertEqual(expected.neg_output_params, sej.neg_output_params)

        self.assertEqual(expected, sej)

    def test_write_odx(self) -> None:
        # Setup jinja environment
        __module_filename = inspect.getsourcefile(odxtools_porsche)
        assert isinstance(__module_filename, str)
        templates_dir = os.path.sep.join([os.path.dirname(__module_filename), "templates"])
        jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(templates_dir))
        jinja_env.globals["odxraise"] = jinja2_odxraise_helper
        jinja_env.globals["make_xml_attrib"] = make_xml_attrib
        jinja_env.globals["make_bool_xml_attrib"] = make_bool_xml_attrib
        jinja_env.globals["getattr"] = getattr
        jinja_env.globals["hasattr"] = hasattr

        # Small template
        template = jinja_env.from_string("""
            {%- import('macros/printSingleEcuJob.xml.jinja2') as psej %}
            {{psej.printSingleEcuJob(singleecujob)}}
        """)

        rawodx: str = template.render(singleecujob=self.singleecujob_object)

        # Remove whitespace
        actual = rawodx.replace(" ", "")
        expected = self.singleecujob_odx.replace(" ", "")

        # Assert equality of outputted string
        self.maxDiff = None
        self.assertEqual(expected, actual)

        # Assert equality of objects
        # This tests the idempotency of read-write
        sej = SingleEcuJob.from_et(
            ElementTree.fromstring(rawodx), OdxDocContext(Version("2.2.0"), doc_frags))
        self.assertEqual(self.singleecujob_object, sej)

    def test_default_lists(self) -> None:
        """Test that empty lists are assigned to list-attributes if no explicit value is passed."""
        sej = SingleEcuJob(
            odx_id=OdxLinkId("ID.SomeID", doc_frags),
            short_name="SN.SomeShortName",
            prog_codes=[ProgCode(
                code_file="abc.jar",
                syntax="abc",
                revision="12.34",
            )],
        )
        self.assertEqual(sej.functional_class_refs, [])
        self.assertEqual(sej.input_params, NamedItemList())
        self.assertEqual(sej.output_params, NamedItemList())
        self.assertEqual(sej.neg_output_params, NamedItemList())
        self.assertEqual(sej.prog_codes[0].library_refs, [])

    def test_resolve_odxlinks(self) -> None:
        ecu_variant_raw = EcuVariantRaw(
            variant_type=DiagLayerType.ECU_VARIANT,
            odx_id=OdxLinkId("ID.bv", doc_frags),
            short_name="bv",
            functional_classes=NamedItemList([self.context.extensiveTask]),
            diag_comms_raw=[self.singleecujob_object],
            additional_audiences=NamedItemList([self.context.specialAudience]),
            libraries=NamedItemList([
                Library(
                    short_name="great_lib",
                    odx_id=OdxLinkId("my.favourite.lib", doc_frags),
                    code_file="great_lib.py",
                    syntax="PYTHON",
                    revision="3.141529",
                    entrypoint="i_am_great")
            ]),
        )
        ecu_variant = EcuVariant(diag_layer_raw=ecu_variant_raw)
        odxlinks = OdxLinkDatabase()
        odxlinks.update(ecu_variant._build_odxlinks())
        # these objects are actually part of the diag layer's
        # diag_data_dictionary_spec, but it is less hassle to
        # "side-load" them...
        odxlinks.update({
            self.context.extensiveTask.odx_id: self.context.extensiveTask,
            self.context.specialAudience.odx_id: self.context.specialAudience,
            self.context.inputDOP.odx_id: self.context.inputDOP,
            self.context.outputDOP.odx_id: self.context.outputDOP,
            self.context.negOutputDOP.odx_id: self.context.negOutputDOP,
        })

        db = Database()
        db.add_auxiliary_file("abc.jar",
                              BytesIO(b"this is supposed to be a JAR archive, but it isn't (HARR)"))
        db.add_auxiliary_file(
            "great_lib.py",
            BytesIO(b"def i_am_great():\n"
                    b"    print('The greatest algorithm eva!')"))

        ecu_variant._resolve_odxlinks(odxlinks)
        ecu_variant._finalize_init(db, odxlinks)

        self.assertEqual(self.context.extensiveTask,
                         self.singleecujob_object.functional_classes.extensiveTask)
        self.assertEqual(self.context.specialAudience,
                         odxrequire(self.singleecujob_object.audience).enabled_audiences[0])

        self.assertEqual(self.context.inputDOP, self.singleecujob_object.input_params[0].dop)
        self.assertEqual(self.context.outputDOP, self.singleecujob_object.output_params[0].dop)
        self.assertEqual(self.context.negOutputDOP,
                         self.singleecujob_object.neg_output_params[0].dop)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .diagcomm import DiagComm
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from .snrefcontext import SnRefContext


@dataclass(kw_only=True)
class TableDiagCommConnector:
    semantic: str

    diag_comm_ref: OdxLinkRef | None = None
    diag_comm_snref: str | None = None

    @property
    def diag_comm(self) -> DiagComm:
        return self._diag_comm

    @staticmethod
    def from_et(et_element: ElementTree.Element,
                context: OdxDocContext) -> "TableDiagCommConnector":

        semantic = odxrequire(et_element.findtext("SEMANTIC"))

        diag_comm_ref = OdxLinkRef.from_et(et_element.find("DIAG-COMM-REF"), context)
        diag_comm_snref = None
        if (dc_snref_elem := et_element.find("DIAG-COMM-SNREF")) is not None:
            diag_comm_snref = odxrequire(dc_snref_elem.get("SHORT-NAME"))

        return TableDiagCommConnector(
            semantic=semantic, diag_comm_ref=diag_comm_ref, diag_comm_snref=diag_comm_snref)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if self.diag_comm_ref is not None:
            self._diag_comm = odxlinks.resolve(self.diag_comm_ref, DiagComm)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        if self.diag_comm_snref is not None:
            dl = odxrequire(context.diag_layer)
            self._diag_comm = resolve_snref(self.diag_comm_snref, dl.diag_comms, DiagComm)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from .companyrevisioninfo import CompanyRevisionInfo
from .exceptions import odxrequire
from .modification import Modification
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .snrefcontext import SnRefContext
from .teammember import TeamMember


@dataclass(kw_only=True)
class DocRevision:
    """
    Representation of a single revision of the relevant object.
    """

    team_member_ref: OdxLinkRef | None = None
    revision_label: str | None = None
    state: str | None = None
    date: str
    tool: str | None = None
    company_revision_infos: list[CompanyRevisionInfo] = field(default_factory=list)
    modifications: list[Modification] = field(default_factory=list)

    @property
    def team_member(self) -> TeamMember | None:
        return self._team_member

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "DocRevision":

        team_member_ref = OdxLinkRef.from_et(et_element.find("TEAM-MEMBER-REF"), context)
        revision_label = et_element.findtext("REVISION-LABEL")
        state = et_element.findtext("STATE")
        date = odxrequire(et_element.findtext("DATE"))
        tool = et_element.findtext("TOOL")

        company_revision_infos = [
            CompanyRevisionInfo.from_et(cri_elem, context)
            for cri_elem in et_element.iterfind("COMPANY-REVISION-INFOS/"
                                                "COMPANY-REVISION-INFO")
        ]

        modifications = [
            Modification.from_et(mod_elem, context)
            for mod_elem in et_element.iterfind("MODIFICATIONS/MODIFICATION")
        ]

        return DocRevision(
            team_member_ref=team_member_ref,
            revision_label=revision_label,
            state=state,
            date=date,
            tool=tool,
            company_revision_infos=company_revision_infos,
            modifications=modifications,
        )

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._team_member: TeamMember | None = None
        if self.team_member_ref is not None:
            self._team_member = odxlinks.resolve(self.team_member_ref, TeamMember)

        for cri in self.company_revision_infos:
            cri._resolve_odxlinks(odxlinks)

        for mod in self.modifications:
            mod._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        for cri in self.company_revision_infos:
            cri._resolve_snrefs(context)

        for mod in self.modifications:
            mod._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
import pytest

from odxtools.utils import is_short_name, is_short_name_path


@pytest.mark.parametrize(
    "val, expected",
    [
        ("abcXYZ", True),
        ("_aA09_", True),
        ("abcde+", False),
        ("+abcde", False),
        ("ab++cd", False),
    ],
)
def test_is_short_name(val: str, expected: bool) -> None:
    assert is_short_name(val) == expected


@pytest.mark.parametrize(
    "val, expected",
    [
        ("abcdefghi", True),
        ("abc.def.g", True),
        (".abcdefgh", False),
        ("abc..defg", False),
        ("abc.defg.", False),
    ],
)
def test_is_short_name_path(val: str, expected: bool) -> None:
    assert is_short_name_path(val) == expected



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .dataobjectproperty import DataObjectProperty
from .exceptions import odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .snrefcontext import SnRefContext


@dataclass(kw_only=True)
class MultiplexerSwitchKey:
    """
    The object that determines the case to be used by a multiplexer
    """
    byte_position: int
    bit_position: int | None = None
    dop_ref: OdxLinkRef

    @property
    def dop(self) -> DataObjectProperty:
        return self._dop

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "MultiplexerSwitchKey":
        byte_position = int(odxrequire(et_element.findtext("BYTE-POSITION")))
        bit_position_str = et_element.findtext("BIT-POSITION")
        bit_position = int(bit_position_str) if bit_position_str is not None else None
        dop_ref = odxrequire(OdxLinkRef.from_et(et_element.find("DATA-OBJECT-PROP-REF"), context))

        return MultiplexerSwitchKey(
            byte_position=byte_position,
            bit_position=bit_position,
            dop_ref=dop_ref,
        )

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._dop = odxlinks.resolve(self.dop_ref, DataObjectProperty)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        pass



#! /usr/bin/python3
#
# SPDX-License-Identifier: MIT
import sys
import time

import can
import isotp

if len(sys.argv) != 5:
    print(f"Usage: {sys.argv[0]} CHANNEL RX_ID TX_ID PAYLOAD_STRING")
    print(f"")
    print(f"Send an ISO-TP telegram over a CAN bus using specified IDs")
    print(f"")
    print(f"Arguments:")
    print(f" CHANNEL         The device name of the CAN network interface (e.g., can0)")
    print(f" RX_ID           The CAN ID used by the peer for receiving (e.g., 0x600)")
    print(f" TX_ID           The CAN ID used by us for flow control (e.g., 0x601)")
    print(f' PAYLOAD_STRING  The string which ought to be send to the peer (e.g., "hello, world!")')
    sys.exit(1)

channel = sys.argv[1]
rx_id = int(sys.argv[2], 0)
tx_id = int(sys.argv[3], 0)
payload = sys.argv[4].encode()


def isotp_error_handler(error: str) -> None:
    print(f"An isotp error occoured: {error}")


isotp_socket = isotp.socket()

# set the ISO-TP flow control options:
#
# stmin: minimum frame separation time [ms]
# bs: maximum block size. (Must be smaller than 4096?)
isotp_socket.set_fc_opts(stmin=5, bs=100)

can_bus = can.Bus(channel=channel, interface="socketcan")

# note that we specify the CAN IDs from the ECU's point of view, i.e.,
# from the tester's (our) perspective, they are reversed.
isotp_addr = isotp.Address(rxid=tx_id, txid=rx_id)

isotp_stack = isotp.CanStack(can_bus, address=isotp_addr, error_handler=isotp_error_handler)

isotp_stack.send(payload)

while isotp_stack.transmitting():
    isotp_stack.process()
    time.sleep(isotp_stack.sleep_time())

print("Payload transmission done.")

can_bus.shutdown()



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any
from xml.etree import ElementTree

from .admindata import AdminData
from .companydata import CompanyData
from .element import IdentifiableElement
from .exceptions import odxrequire
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import DocType, OdxDocFragment, OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext
from .specialdatagroup import SpecialDataGroup
from .utils import dataclass_fields_asdict

if TYPE_CHECKING:
    from .database import Database


@dataclass(kw_only=True)
class OdxCategory(IdentifiableElement):
    """This is the base class for all top-level container classes in ODX"""

    admin_data: AdminData | None = None
    company_datas: NamedItemList[CompanyData] = field(default_factory=NamedItemList)
    sdgs: list[SpecialDataGroup] = field(default_factory=list)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "OdxCategory":
        raise Exception("Calling `._from_et()` is not allowed for OdxCategory. "
                        "Use `OdxCategory.category_from_et()`!")

    @staticmethod
    def category_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                         doc_type: DocType) -> "OdxCategory":

        short_name = odxrequire(et_element.findtext("SHORT-NAME"))
        # create the current ODX "document fragment" (description of the
        # current document for references and IDs)
        context.doc_fragments.append(OdxDocFragment(short_name, doc_type))
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        admin_data = AdminData.from_et(et_element.find("ADMIN-DATA"), context)
        company_datas = NamedItemList([
            CompanyData.from_et(cde, context)
            for cde in et_element.iterfind("COMPANY-DATAS/COMPANY-DATA")
        ])
        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]

        return OdxCategory(admin_data=admin_data, company_datas=company_datas, sdgs=sdgs, **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {self.odx_id: self}

        if self.admin_data is not None:
            result.update(self.admin_data._build_odxlinks())
        for cd in self.company_datas:
            result.update(cd._build_odxlinks())
        for sdg in self.sdgs:
            result.update(sdg._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if self.admin_data is not None:
            self.admin_data._resolve_odxlinks(odxlinks)
        for cd in self.company_datas:
            cd._resolve_odxlinks(odxlinks)
        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

    def _finalize_init(self, database: "Database", odxlinks: OdxLinkDatabase) -> None:
        pass

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        if self.admin_data is not None:
            self.admin_data._resolve_snrefs(context)
        for cd in self.company_datas:
            cd._resolve_snrefs(context)
        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
import warnings
from dataclasses import dataclass
from typing import Any
from xml.etree import ElementTree

from .commrelationvaluetype import CommRelationValueType
from .description import Description
from .diagcomm import DiagComm
from .diagservice import DiagService
from .exceptions import OdxWarning, odxraise, odxrequire
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef, resolve_snref
from .parameters.parameter import Parameter
from .snrefcontext import SnRefContext


@dataclass(kw_only=True)
class CommRelation:
    description: Description | None = None
    relation_type: str
    diag_comm_ref: OdxLinkRef | None = None
    diag_comm_snref: str | None = None
    in_param_if_snref: str | None = None
    #in_param_if_snpathref: Optional[str] = None # TODO
    out_param_if_snref: str | None = None
    #out_param_if_snpathref: Optional[str] = None # TODO
    value_type_raw: CommRelationValueType | None = None

    @property
    def diag_comm(self) -> DiagComm:
        return self._diag_comm

    @property
    def in_param_if(self) -> Parameter | None:
        return self._in_param_if

    @property
    def out_param_if(self) -> Parameter | None:
        return self._out_param_if

    @property
    def value_type(self) -> CommRelationValueType:
        if self.value_type_raw is None:
            return CommRelationValueType.CURRENT

        return self.value_type_raw

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "CommRelation":
        description = Description.from_et(et_element.find("DESC"), context)
        relation_type = odxrequire(et_element.findtext("RELATION-TYPE"))

        diag_comm_ref = OdxLinkRef.from_et(et_element.find("DIAG-COMM-REF"), context)
        diag_comm_snref = None
        if (diag_comm_snref_elem := et_element.find("DIAG-COMM-SNREF")) is not None:
            diag_comm_snref = odxrequire(diag_comm_snref_elem.get("SHORT-NAME"))

        in_param_if_snref = None
        if (in_param_if_snref_elem := et_element.find("IN-PARAM-IF-SNREF")) is not None:
            in_param_if_snref = odxrequire(in_param_if_snref_elem.get("SHORT-NAME"))

        if et_element.find("IN-PARAM-IF-SNPATHREF") is not None:
            warnings.warn("SNPATHREFs are not supported by odxtools yet", OdxWarning, stacklevel=1)

        out_param_if_snref = None
        if (out_param_if_snref_elem := et_element.find("OUT-PARAM-IF-SNREF")) is not None:
            out_param_if_snref = odxrequire(out_param_if_snref_elem.get("SHORT-NAME"))

        if et_element.find("OUT-PARAM-IF-SNPATHREF") is not None:
            warnings.warn("SNPATHREFs are not supported by odxtools yet", OdxWarning, stacklevel=1)

        value_type_raw = None
        if (value_type_str := et_element.attrib.get("VALUE-TYPE")) is not None:
            try:
                value_type_raw = CommRelationValueType(value_type_str)
            except ValueError:
                odxraise(f"Encountered unknown comm relation value type '{value_type_str}'")

        return CommRelation(
            description=description,
            relation_type=relation_type,
            diag_comm_ref=diag_comm_ref,
            diag_comm_snref=diag_comm_snref,
            in_param_if_snref=in_param_if_snref,
            out_param_if_snref=out_param_if_snref,
            value_type_raw=value_type_raw)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if self.diag_comm_ref is not None:
            self._diag_comm = odxlinks.resolve(self.diag_comm_ref, DiagComm)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        diag_layer = odxrequire(context.diag_layer)

        if self.diag_comm_snref is not None:
            self._diag_comm = resolve_snref(self.diag_comm_snref, diag_layer.diag_comms, DiagComm)

        service = self.diag_comm
        if not isinstance(service, DiagService):
            odxraise(f"DIAG-VARIABLE references non-service {type(service).__name__} "
                     f"diagnostic communication")
            return

        self._in_param_if = None
        if self.in_param_if_snref is not None:
            self._in_param_if = resolve_snref(self.in_param_if_snref,
                                              odxrequire(service.request).parameters, Parameter)

        self._out_param_if = None
        if self.out_param_if_snref is not None:
            self._out_param_if = resolve_snref(self.out_param_if_snref,
                                               odxrequire(service.positive_responses[0]).parameters,
                                               Parameter)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from xml.etree import ElementTree

from ..exceptions import DecodeError, EncodeError, odxassert, odxraise, odxrequire
from ..odxdoccontext import OdxDocContext
from ..odxtypes import AtomicOdxType, DataType
from ..utils import dataclass_fields_asdict
from .compucategory import CompuCategory
from .compumethod import CompuMethod
from .intervaltype import IntervalType
from .limit import Limit


@dataclass(kw_only=True)
class TabIntpCompuMethod(CompuMethod):
    """A table-based interpolated compu method provides a continuous
    transfer function based on piecewise linear interpolation.

    A `TabIntpCompuMethod` is defined by a set of points. Each point
    is an (internal, physical) value pair.  When converting from
    internal to physical or vice-versa, the result is linearly
    interpolated.

    The function defined by a `TabIntpCompuMethod` is similar to the
    one of a `ScaleLinearCompuMethod` with the following differences:

    * `TabIntpCompuMethod`s are always continuous whereas
      `ScaleLinearCompuMethod` might exhibit gaps
    * `TabIntpCompuMethod`s are always invertible: Even if the linear
      interpolation is not monotonic, the first matching interval is
      used.

    For details, refer to ASAM specification MCD-2 D (ODX), section 7.3.6.6.8.

    """

    @property
    def internal_points(self) -> list[float | int]:
        return self._internal_points

    @property
    def physical_points(self) -> list[float | int]:
        return self._physical_points

    @property
    def internal_lower_limit(self) -> Limit:
        return self._internal_lower_limit

    @property
    def internal_upper_limit(self) -> Limit:
        return self._internal_upper_limit

    @property
    def physical_lower_limit(self) -> Limit:
        return self._physical_lower_limit

    @property
    def physical_upper_limit(self) -> Limit:
        return self._physical_upper_limit

    @staticmethod
    def compu_method_from_et(et_element: ElementTree.Element, context: OdxDocContext, *,
                             internal_type: DataType,
                             physical_type: DataType) -> "TabIntpCompuMethod":
        cm = CompuMethod.compu_method_from_et(
            et_element, context, internal_type=internal_type, physical_type=physical_type)
        kwargs = dataclass_fields_asdict(cm)

        return TabIntpCompuMethod(**kwargs)

    def __post_init__(self) -> None:
        odxassert(self.category == CompuCategory.TAB_INTP,
                  "TabIntpCompuMethod must exibit TAB-INTP category")

        self._internal_points: list[int | float] = []
        self._physical_points: list[int | float] = []
        for scale in odxrequire(self.compu_internal_to_phys).compu_scales:
            internal_point = odxrequire(scale.lower_limit).value
            physical_point = odxrequire(scale.compu_const).value

            if not isinstance(internal_point, (float, int)):
                odxraise("The type of values of tab-intp compumethods must "
                         "either int or float")
            if not isinstance(physical_point, (float, int)):
                odxraise("The type of values of tab-intp compumethods must "
                         "either int or float")

            self._internal_points.append(internal_point)
            self._physical_points.append(physical_point)

        self._physical_lower_limit = Limit(
            value_raw=str(min(self._physical_points)),
            value_type=self.physical_type,
            interval_type=IntervalType.CLOSED)
        self._physical_upper_limit = Limit(
            value_raw=str(max(self._physical_points)),
            value_type=self.physical_type,
            interval_type=IntervalType.CLOSED)

        self._internal_lower_limit = Limit(
            value_raw=str(min(self._internal_points)),
            value_type=self.internal_type,
            interval_type=IntervalType.CLOSED)
        self._internal_upper_limit = Limit(
            value_raw=str(max(self._internal_points)),
            value_type=self.internal_type,
            interval_type=IntervalType.CLOSED)

        self.__assert_validity()

    def __assert_validity(self) -> None:
        odxassert(len(self.internal_points) == len(self.physical_points))

        odxassert(
            self.internal_type in [
                DataType.A_INT32,
                DataType.A_UINT32,
                DataType.A_FLOAT32,
                DataType.A_FLOAT64,
            ], "Internal data type of TAB-INTP compumethod must be one of"
            " [A_INT32, A_UINT32, A_FLOAT32, A_FLOAT64]")
        odxassert(
            self.physical_type in [
                DataType.A_INT32,
                DataType.A_UINT32,
                DataType.A_FLOAT32,
                DataType.A_FLOAT64,
            ], "Physical data type of TAB-INTP compumethod must be one of"
            " [A_INT32, A_UINT32, A_FLOAT32, A_FLOAT64]")

    def __piecewise_linear_interpolate(self, x: int | float, range_samples: list[int | float],
                                       domain_samples: list[int | float]) -> float | None:
        for i in range(0, len(range_samples) - 1):
            if (x0 := range_samples[i]) <= x and x <= (x1 := range_samples[i + 1]):
                y0 = domain_samples[i]
                y1 = domain_samples[i + 1]
                return y0 + (x - x0) * (y1 - y0) / (x1 - x0)

        return None

    def convert_physical_to_internal(self, physical_value: AtomicOdxType) -> AtomicOdxType:
        if not isinstance(physical_value, (int, float)):
            odxraise("The type of values of tab-intp compumethods must "
                     "either int or float", EncodeError)
            return None

        odxassert(
            isinstance(physical_value, (int, float)),
            "Only integers and floats can be piecewise linearly interpolated", EncodeError)
        result = self.__piecewise_linear_interpolate(physical_value, self._physical_points,
                                                     self._internal_points)

        if result is None:
            odxraise(
                f"Internal value {physical_value!r} must be inside the range"
                f" [{min(self.physical_points)}, {max(self.physical_points)}]", EncodeError)

        res = self.internal_type.make_from(result)

        return res

    def convert_internal_to_physical(self, internal_value: AtomicOdxType) -> AtomicOdxType:
        if not isinstance(internal_value, (int, float)):
            odxraise(
                "The internal type of values of tab-intp compumethods must "
                "either int or float", EncodeError)
            return None

        odxassert(
            isinstance(internal_value, (int, float)),
            "Only integers and floats can be piecewise linearly interpolated", DecodeError)

        result = self.__piecewise_linear_interpolate(internal_value, self._internal_points,
                                                     self._physical_points)

        if result is None:
            odxraise(
                f"Internal value {internal_value!r} must be inside the range"
                f" [{min(self.internal_points)}, {max(self.internal_points)}]", DecodeError)
            return None

        res = self.physical_type.make_from(result)

        return res

    def is_valid_physical_value(self, physical_value: AtomicOdxType) -> bool:
        if not isinstance(physical_value, (int, float)):
            return False

        return min(self.physical_points) <= physical_value and physical_value <= max(
            self.physical_points)

    def is_valid_internal_value(self, internal_value: AtomicOdxType) -> bool:
        if not isinstance(internal_value, (int, float)):
            return False

        return min(self.internal_points) <= internal_value and internal_value <= max(
            self.internal_points)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any
from xml.etree import ElementTree

from ..diagvariable import DiagVariable
from ..exceptions import odxraise
from ..nameditemlist import NamedItemList
from ..odxdoccontext import OdxDocContext
from ..odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from ..parentref import ParentRef
from ..snrefcontext import SnRefContext
from ..utils import dataclass_fields_asdict
from ..variablegroup import VariableGroup
from .hierarchyelementraw import HierarchyElementRaw


@dataclass(kw_only=True)
class FunctionalGroupRaw(HierarchyElementRaw):
    """This is a diagnostic layer for common functionality of an ECU
    """

    diag_variables_raw: list[DiagVariable | OdxLinkRef] = field(default_factory=list)
    variable_groups: NamedItemList[VariableGroup] = field(default_factory=NamedItemList)
    parent_refs: list[ParentRef] = field(default_factory=list)

    @property
    def diag_variables(self) -> NamedItemList[DiagVariable]:
        return self._diag_variables

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "FunctionalGroupRaw":
        # objects contained by diagnostic layers exibit an additional
        # document fragment for the diag layer, so we use the document
        # fragments of the odx id of the diag layer for IDs of
        # contained objects.
        her = HierarchyElementRaw.from_et(et_element, context)
        kwargs = dataclass_fields_asdict(her)

        diag_variables_raw: list[DiagVariable | OdxLinkRef] = []
        if (dv_elems := et_element.find("DIAG-VARIABLES")) is not None:
            for dv_proxy_elem in dv_elems:
                dv_proxy: OdxLinkRef | DiagVariable
                if dv_proxy_elem.tag == "DIAG-VARIABLE-REF":
                    dv_proxy = OdxLinkRef.from_et(dv_proxy_elem, context)
                elif dv_proxy_elem.tag == "DIAG-VARIABLE":
                    dv_proxy = DiagVariable.from_et(dv_proxy_elem, context)
                else:
                    odxraise()

                diag_variables_raw.append(dv_proxy)

        variable_groups = NamedItemList([
            VariableGroup.from_et(vg_elem, context)
            for vg_elem in et_element.iterfind("VARIABLE-GROUPS/VARIABLE-GROUP")
        ])

        parent_refs = [
            ParentRef.from_et(pr_elem, context)
            for pr_elem in et_element.iterfind("PARENT-REFS/PARENT-REF")
        ]

        return FunctionalGroupRaw(
            diag_variables_raw=diag_variables_raw,
            variable_groups=variable_groups,
            parent_refs=parent_refs,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = super()._build_odxlinks()

        for dv_proxy in self.diag_variables_raw:
            if not isinstance(dv_proxy, OdxLinkRef):
                result.update(dv_proxy._build_odxlinks())

        for parent_ref in self.parent_refs:
            result.update(parent_ref._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        self._diag_variables: NamedItemList[DiagVariable] = NamedItemList()
        for dv_proxy in self.diag_variables_raw:
            if isinstance(dv_proxy, OdxLinkRef):
                dv = odxlinks.resolve(dv_proxy, DiagVariable)
            else:
                dv_proxy._resolve_odxlinks(odxlinks)
                dv = dv_proxy

            self._diag_variables.append(dv)

        for parent_ref in self.parent_refs:
            parent_ref._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

        for dv_proxy in self.diag_variables_raw:
            if not isinstance(dv_proxy, OdxLinkRef):
                dv_proxy._resolve_snrefs(context)

        for parent_ref in self.parent_refs:
            parent_ref._resolve_snrefs(context)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, cast
from xml.etree import ElementTree

from .admindata import AdminData
from .compositecodec import (composite_codec_decode_from_pdu, composite_codec_encode_into_pdu,
                             composite_codec_get_coded_const_prefix,
                             composite_codec_get_free_parameters,
                             composite_codec_get_required_parameters,
                             composite_codec_get_static_bit_length)
from .decodestate import DecodeState
from .element import IdentifiableElement
from .encodestate import EncodeState
from .exceptions import odxraise
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId
from .odxtypes import ParameterValue, ParameterValueDict
from .parameters.createanyparameter import create_any_parameter_from_et
from .parameters.parameter import Parameter
from .snrefcontext import SnRefContext
from .specialdatagroup import SpecialDataGroup
from .utils import dataclass_fields_asdict


class ResponseType(Enum):
    POSITIVE = "POS-RESPONSE"
    NEGATIVE = "NEG-RESPONSE"
    GLOBAL_NEGATIVE = "GLOBAL-NEG-RESPONSE"


@dataclass(kw_only=True)
class Response(IdentifiableElement):
    """Represents all information related to an UDS response

    This class implements the `CompositeCodec` interface.
    """

    response_type: ResponseType

    admin_data: AdminData | None = None
    parameters: NamedItemList[Parameter] = field(default_factory=NamedItemList)
    sdgs: list[SpecialDataGroup] = field(default_factory=list)

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "Response":
        """Reads a response."""
        kwargs = dataclass_fields_asdict(IdentifiableElement.from_et(et_element, context))

        try:
            response_type = ResponseType(et_element.tag)
        except ValueError:
            response_type = cast(ResponseType, None)
            odxraise(f"Encountered unknown response type '{et_element.tag}'")

        admin_data = AdminData.from_et(et_element.find("ADMIN-DATA"), context)
        parameters = NamedItemList([
            create_any_parameter_from_et(et_parameter, context)
            for et_parameter in et_element.iterfind("PARAMS/PARAM")
        ])
        sdgs = [SpecialDataGroup.from_et(sdge, context) for sdge in et_element.iterfind("SDGS/SDG")]

        return Response(
            response_type=response_type,
            admin_data=admin_data,
            parameters=parameters,
            sdgs=sdgs,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        result = {self.odx_id: self}

        if self.admin_data is not None:
            result.update(self.admin_data._build_odxlinks())

        for param in self.parameters:
            result.update(param._build_odxlinks())

        for sdg in self.sdgs:
            result.update(sdg._build_odxlinks())

        return result

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        if self.admin_data is not None:
            self.admin_data._resolve_odxlinks(odxlinks)

        for param in self.parameters:
            param._resolve_odxlinks(odxlinks)

        for sdg in self.sdgs:
            sdg._resolve_odxlinks(odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        context.response = self
        context.parameters = self.parameters

        if self.admin_data is not None:
            self.admin_data._resolve_snrefs(context)

        for param in self.parameters:
            param._resolve_snrefs(context)

        for sdg in self.sdgs:
            sdg._resolve_snrefs(context)

        context.response = None
        context.parameters = None

    def encode(self, coded_request: bytes | None = None, **kwargs: ParameterValue) -> bytearray:
        encode_state = EncodeState(triggering_request=coded_request, is_end_of_pdu=True)

        self.encode_into_pdu(physical_value=kwargs, encode_state=encode_state)

        return encode_state.coded_message

    def decode(self, message: bytes) -> ParameterValueDict:
        decode_state = DecodeState(coded_message=message)
        param_values = self.decode_from_pdu(decode_state)

        if not isinstance(param_values, dict):
            odxraise("Decoding a response must result in a dictionary")

        return cast(ParameterValueDict, param_values)

    def encode_into_pdu(self, physical_value: ParameterValue | None,
                        encode_state: EncodeState) -> None:
        composite_codec_encode_into_pdu(self, physical_value, encode_state)

    def decode_from_pdu(self, decode_state: DecodeState) -> ParameterValue:
        return composite_codec_decode_from_pdu(self, decode_state)

    def get_static_bit_length(self) -> int | None:
        return composite_codec_get_static_bit_length(self)

    @property
    def required_parameters(self) -> list[Parameter]:
        return composite_codec_get_required_parameters(self)

    @property
    def free_parameters(self) -> list[Parameter]:
        return composite_codec_get_free_parameters(self)

    def print_free_parameters_info(self) -> None:
        """Return a human readable description of the structure's
        free parameters.
        """
        from .parameterinfo import parameter_info

        print(parameter_info(self.free_parameters), end="")

    def coded_const_prefix(self, request_prefix: bytes = b'') -> bytes:
        return composite_codec_get_coded_const_prefix(self, request_prefix)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import Any, cast
from xml.etree import ElementTree

from .exceptions import odxraise, odxrequire
from .library import Library
from .nameditemlist import NamedItemList
from .odxdoccontext import OdxDocContext
from .odxlink import OdxLinkDatabase, OdxLinkId, OdxLinkRef
from .snrefcontext import SnRefContext


@dataclass(kw_only=True)
class ProgCode:
    """A reference to code that is executed by a single ECU job"""
    code_file: str
    encryption: str | None = None
    syntax: str
    revision: str
    entrypoint: str | None = None
    library_refs: list[OdxLinkRef] = field(default_factory=list)

    @property
    def code(self) -> bytes:
        return self._code

    @property
    def libraries(self) -> NamedItemList[Library]:
        return self._libraries

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "ProgCode":
        code_file = odxrequire(et_element.findtext("CODE-FILE"))
        encryption = et_element.findtext("ENCRYPTION")
        syntax = odxrequire(et_element.findtext("SYNTAX"))
        revision = odxrequire(et_element.findtext("REVISION"))
        entrypoint = et_element.findtext("ENTRYPOINT")

        library_refs = [
            odxrequire(OdxLinkRef.from_et(el, context))
            for el in et_element.iterfind("LIBRARY-REFS/LIBRARY-REF")
        ]

        return ProgCode(
            code_file=code_file,
            encryption=encryption,
            syntax=syntax,
            revision=revision,
            entrypoint=entrypoint,
            library_refs=library_refs,
        )

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        return {}

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        self._libraries = NamedItemList([odxlinks.resolve(x, Library) for x in self.library_refs])

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        aux_file = odxrequire(context.database).auxiliary_files.get(self.code_file)

        if aux_file is None:
            odxraise(f"Reference to auxiliary file '{self.code_file}' "
                     f"could not be resolved")
            self._code: bytes = cast(bytes, None)
            return

        self._code = aux_file.read()
        aux_file.seek(0)



# SPDX-License-Identifier: MIT
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any
from xml.etree import ElementTree

from .comparam import Comparam
from .complexcomparam import ComplexComparam
from .dataobjectproperty import DataObjectProperty
from .nameditemlist import NamedItemList
from .odxcategory import OdxCategory
from .odxdoccontext import OdxDocContext
from .odxlink import DocType, OdxLinkDatabase, OdxLinkId
from .snrefcontext import SnRefContext
from .unitspec import UnitSpec
from .utils import dataclass_fields_asdict

if TYPE_CHECKING:
    from .database import Database


@dataclass(kw_only=True)
class ComparamSubset(OdxCategory):
    comparams: NamedItemList[Comparam] = field(default_factory=NamedItemList)
    complex_comparams: NamedItemList[ComplexComparam] = field(default_factory=NamedItemList)
    data_object_props: NamedItemList[DataObjectProperty] = field(default_factory=NamedItemList)
    unit_spec: UnitSpec | None = None
    category: str | None  # mandatory in ODX 2.2, but non-existent in ODX 2.0

    @staticmethod
    def from_et(et_element: ElementTree.Element, context: OdxDocContext) -> "ComparamSubset":

        category_attrib = et_element.attrib.get("CATEGORY")

        # In ODX 2.0, COMPARAM-SPEC is used, whereas in ODX 2.2, it
        # refers to something else and has been replaced by
        # COMPARAM-SUBSET.
        # - If the 'CATEGORY' attribute is missing (ODX 2.0), use
        #   COMPARAM_SPEC,
        # - else (ODX 2.2), use COMPARAM_SUBSET.
        doc_type = DocType.COMPARAM_SUBSET if category_attrib is not None else DocType.COMPARAM_SPEC
        base_obj = OdxCategory.category_from_et(et_element, context, doc_type=doc_type)
        kwargs = dataclass_fields_asdict(base_obj)

        comparams = NamedItemList(
            [Comparam.from_et(el, context) for el in et_element.iterfind("COMPARAMS/COMPARAM")])
        complex_comparams = NamedItemList([
            ComplexComparam.from_et(el, context)
            for el in et_element.iterfind("COMPLEX-COMPARAMS/COMPLEX-COMPARAM")
        ])
        data_object_props = NamedItemList([
            DataObjectProperty.from_et(el, context)
            for el in et_element.iterfind("DATA-OBJECT-PROPS/DATA-OBJECT-PROP")
        ])
        unit_spec = None
        if (unit_spec_elem := et_element.find("UNIT-SPEC")) is not None:
            unit_spec = UnitSpec.from_et(unit_spec_elem, context)

        return ComparamSubset(
            category=category_attrib,
            comparams=comparams,
            complex_comparams=complex_comparams,
            data_object_props=data_object_props,
            unit_spec=unit_spec,
            **kwargs)

    def _build_odxlinks(self) -> dict[OdxLinkId, Any]:
        odxlinks = super()._build_odxlinks()

        for comparam in self.comparams:
            odxlinks.update(comparam._build_odxlinks())

        for ccomparam in self.complex_comparams:
            odxlinks.update(ccomparam._build_odxlinks())

        for dop in self.data_object_props:
            odxlinks[dop.odx_id] = dop

        if self.unit_spec:
            odxlinks.update(self.unit_spec._build_odxlinks())

        return odxlinks

    def _resolve_odxlinks(self, odxlinks: OdxLinkDatabase) -> None:
        super()._resolve_odxlinks(odxlinks)

        for comparam in self.comparams:
            comparam._resolve_odxlinks(odxlinks)

        for ccomparam in self.complex_comparams:
            ccomparam._resolve_odxlinks(odxlinks)

        for dop in self.data_object_props:
            dop._resolve_odxlinks(odxlinks)

        if self.unit_spec:
            self.unit_spec._resolve_odxlinks(odxlinks)

    def _finalize_init(self, database: "Database", odxlinks: OdxLinkDatabase) -> None:
        super()._finalize_init(database, odxlinks)

    def _resolve_snrefs(self, context: SnRefContext) -> None:
        super()._resolve_snrefs(context)

        for comparam in self.comparams:
            comparam._resolve_snrefs(context)

        for ccomparam in self.complex_comparams:
            ccomparam._resolve_snrefs(context)

        for dop in self.data_object_props:
            dop._resolve_snrefs(context)

        if self.unit_spec:
            self.unit_spec._resolve_snrefs(context)



#! /usr/bin/python3
#
# SPDX-License-Identifier: MIT
#
# This script demonstrates how an existing ODX database can be
# modified programatically. Note that this is pretty hacky...
import argparse
from copy import deepcopy
from typing import TypeVar

import odxtools_porsche
import odxtools_porsche.uds as uds
from examples import somersaultecu
from odxtools.description import Description
from odxtools.diaglayers.diaglayer import DiagLayer
from odxtools.diaglayers.ecuvariant import EcuVariant
from odxtools.diagservice import DiagService
from odxtools.nameditemlist import NamedItemList
from odxtools.odxlink import OdxLinkId, OdxLinkRef
from odxtools.parameters.codedconstparameter import CodedConstParameter
from odxtools.parameters.valueparameter import ValueParameter
from odxtools.request import Request
from odxtools.response import Response, ResponseType

T = TypeVar("T")


def find_named_object(item_list: list[T], name: str) -> T:
    for x in item_list:
        if getattr(x, "short_name", None) == name:
            return x

    raise KeyError(str(name))


FLIC_FLAC_SID = 0xBE

db = somersaultecu.database

dlc = db.diag_layer_containers[0]

# modify the positive response of the tester_present service for all ECUs
somersault_dlr = dlc.base_variants.somersault.diag_layer_raw

# rename the "session_start" and "session_stop" services to
# "start_session" and "stop_session"
start_service = find_named_object(somersault_dlr.diag_comms_raw, "session_start")
assert isinstance(start_service, DiagService)
start_service.short_name = "start_session"

stop_service = find_named_object(somersault_dlr.diag_comms_raw, "session_stop")
assert isinstance(stop_service, DiagService)
stop_service.short_name = "stop_session"

tester_present_service = find_named_object(somersault_dlr.diag_comms_raw, "tester_present")
assert isinstance(tester_present_service, DiagService)
tester_present_pr = tester_present_service.positive_responses[0]
param = tester_present_pr.parameters.status
assert isinstance(param, ValueParameter)
param.dop_ref = OdxLinkRef.from_id(somersaultecu.somersault_dops["uint16"].odx_id)

# change the DOP used by the "can_do_backward_flips" parameter of the
# positive response to the "session" service from "bool" to "uint8"
cdbf_param = start_service.positive_responses[0].parameters.can_do_backward_flips
doc_frags = start_service.odx_id.doc_fragments
assert isinstance(cdbf_param, ValueParameter)
cdbf_param.dop_ref = OdxLinkRef("somersault.DOP.uint8", doc_frags)

# change the byte and bit positions of the "num_flips_done" parameter
# of the "grudging_backward" response for all variants variant
gb_response = find_named_object(somersault_dlr.positive_responses, "grudging_backward")
param = gb_response.parameters.num_flips_done
param.byte_position = 2
param.bit_position = 4

# add a new "somersault_young" variant which can do flic-flacs and
# does not take any instructions
somersault_lazy = dlc.ecu_variants.somersault_lazy
somersault_young_dlr = deepcopy(somersault_lazy.ecu_variant_raw)
somersault_young_dlr.short_name = "somersault_young"
somersault_young_dlr.odx_id = OdxLinkId("ECU.somersault_young",
                                        somersault_young_dlr.odx_id.doc_fragments)
somersault_young_dlr.description = \
Description.from_string("""<p>A young version of the somersault ECU

It is as grumpy as the lazy variant, but it is more agile, so it can do flic-flacs.

On the flipside, it is unwilling to take any instructions, so no
operational parameters can be set. Finally, it is unwilling to compete
(i.e. it does not time its somersaults).</p>""")
somersault_young = DiagLayer(diag_layer_raw=somersault_young_dlr)

# remove the "sault_time" parameter from the positive response of the
# "do_forward_flips" service.
do_forward_flips_service = deepcopy(somersault_lazy.services.do_forward_flips)
somersault_young_dlr.diag_comms_raw.append(do_forward_flips_service)
pr = do_forward_flips_service.positive_responses[0]
new_params = [x for x in pr.parameters if getattr(x, "short_name", None) != "sault_time"]
pr.parameters = NamedItemList(new_params)

# add a new "flic-flac" service
flic_flac_request = Request(
    odx_id=OdxLinkId("somersault.RQ.flic_flac", doc_frags),
    short_name="RQ_flic_flac",
    parameters=NamedItemList([
        CodedConstParameter(
            short_name="sid",
            diag_coded_type=somersaultecu.somersault_diagcodedtypes["uint8"],
            byte_position=0,
            coded_value_raw=str(FLIC_FLAC_SID),
        )
    ]),
)
somersault_young_dlr.requests.append(flic_flac_request)

flic_flac_positive_response = Response(
    odx_id=OdxLinkId("somersault.PR.flic_flac", doc_frags),
    short_name="PR_flic_flac",
    response_type=ResponseType.POSITIVE,
    parameters=NamedItemList([
        CodedConstParameter(
            short_name="sid",
            diag_coded_type=somersaultecu.somersault_diagcodedtypes["uint8"],
            byte_position=0,
            coded_value_raw=str(uds.positive_response_id(FLIC_FLAC_SID)),
        ),
        ValueParameter(
            short_name="can_do_backward_flips",
            byte_position=1,
            dop_ref=OdxLinkRef("somersault.DOP.boolean", doc_frags),
        ),
    ]),
)
somersault_young_dlr.positive_responses.append(flic_flac_positive_response)

flic_flac_service = DiagService(
    odx_id=OdxLinkId("somersault.service.flic_flac", doc_frags),
    short_name="flic_flac",
    semantic="FUNCTION",
    request_ref=OdxLinkRef.from_id(flic_flac_request.odx_id),
    pos_response_refs=[
        OdxLinkRef.from_id(flic_flac_positive_response.odx_id),
    ],
    neg_response_refs=[
        OdxLinkRef.from_id(somersaultecu.somersault_negative_responses["general"].odx_id),
    ],
)

# create a new list of diagnostic communications that does not include
# the "set_operation_params" and "compulsory_program" services
ss_young_diag_comms_raw = [
    x for x in somersault_young_dlr.diag_comms_raw
    if getattr(x, "short_name", None) not in ("set_operation_params", "compulsory_program")
]

# append the flic-flac service
ss_young_diag_comms_raw.append(flic_flac_service)

# change the list of the ECU's diag comms
somersault_young_dlr.diag_comms_raw = ss_young_diag_comms_raw

dlc.ecu_variants.append(EcuVariant(diag_layer_raw=somersault_young_dlr))

# make the database consistent. Note: For just writing to disk this is
# not necessary (but it is useful if the database is going to be used
# for something else later...)
db.refresh()

if __name__ == "__main__":
    argparser = argparse.ArgumentParser(
        description="\n".join([
            "Creates a simple sample PDX file for a modified 'somersault' ECU from scratch.", "",
            "The modified PDX file is primarily intended to be used a demo for the ",
            "'compare' command line tool."
        ]),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    argparser.add_argument(
        "output_pdx_file",
        metavar="OUTPUT_PDX_FILE",
        help="Path to the where the resulting .pdx file is written",
    )

    args = argparser.parse_args()

    # write the result
    odxtools.write_pdx_file(args.output_pdx_file, db)
